program GCON;

// Version 1.18 vom 30.09.2013


// DEVICE = mega1284P;

// {$I mega1284_Register.INC}



DEVICE = mega644P;

{$I mega644P_Register.INC}


CONST

  BAUDRATE : WORD = 25253;

  // ((ProcClock+BAUDRATE*8)/(BAUDRATE*16)-1)
  // (((25253*8)+20000000)/(25253*16))-1) = ((202024+20000000)/404048)-1 =
  // (20202024/404048)-1 = 50-1 = 49
  UBRR_VAL : WORD = 49;

  XON  : Byte = $11;
  XOFF : Byte = $13;


  // Werte fuer Font_Bez
  fnt_6x8 : Byte = 1;
  fnt_8x8 : Byte = 2;



  // Festwerte fuer Farben
  cl_White    : Byte = $FF; // %11111111
  cl_Black    : Byte = $00; // %00000000
  cl_Red      : Byte = $43; // %11000011
  cl_Green    : Byte = $4C; // %11001100
  cl_Blue     : Byte = $B0; // %11110000
  cl_Yellow   : Byte = $4F; // %11001100
  cl_Purple   : Byte = $73; // %11001100
  cl_Aqua     : Byte = $7C; // %11001100
  cl_Gray     : Byte = $6A; // %01101010
  cl_DarkGray : Byte = $2A; // %00101010


  // TerminalMode
  Normal_TM            : Byte = 0;
  IBM_ESC_TM           : Byte = 1;
  IBM_CSI_TM           : Byte = 2;
  IBM_CSI_TM_Par_1     : Byte = 3;
  IBM_CSI_TM_Par_2     : Byte = 4;
  IBM_CSI_TM_Par_3     : Byte = 5;
  IBM_GRAFIK_Par_1_TM  : Byte = 6;
  IBM_GRAFIK_Par_2_TM  : Byte = 7;


// Font fuer 6x8 Zeichen
{$I font6x8.INC}

// Font fuer 8x8 Zeichen
{$I font8x8.INC}

// Bild 128x240
{$I BILD5_256.INC}

// Kreis 26x16
{$I KREIS2A.INC}

// Sinus-Tabelle
{$I SinTab.INC}


{--------------------------------------------------------------}

var

  // Globale Variablen

  A16[PortD, 2]     : bit; // A
  WE[PortD, 3]      : bit; // A
  OE[PortD, 4]      : bit; // A
  VS1_GS2[PortD, 5] : bit; // A
  DIR[PortD, 6]     : bit; // A
  LED_GN[PortD, 7]  : bit; // A

 
  FCOLOR, BCOLOR, PIXCOLOR,
  Zeichen : Byte;

  X_KUR_POS, Y_KUR_POS,
  X_PIX_POS, Y_PIX_POS,
  X_END_POS, Y_END_POS,
  Kursor_Schleife: Word;

  Kursor_sichtbar, Kursor_Ein,
  KurArr_Aktuell : Boolean;

  Kursor_Array : Array [0..63] of Byte;  // Breite*Hoehe = 8x8=64
  CHAR_Array   : Array [0..63] of Byte;

  // Variablen zum Icon
  EBENE1_Array : Array [0..415] of Byte;  // Breite*Hoehe = 26x16=416
  E1_X_PIX_POS, E1_Y_PIX_POS : Word;
  EBENE2_Array : Array [0..415] of Byte;  // Breite*Hoehe = 26x16=416
  E2_X_PIX_POS, E2_Y_PIX_POS : Word;
  ICON_Adress : Adress;

  TermMode : Byte;

  CSI_Parameter1, CSI_Parameter2,
  CSI_Parameter3 : Word;

  delta_x, delta_y : Word;

  Sin_dw1, Sin_dw2 : DWord;

  Escape_Sequenz : String[10];


  // Variablen zum Font
  Zeichenbreite, Zeichenhoehe,
  max_Spalten2, max_Zeilen2,
  X_ANF, Y_ANF,
  max_Spalten, max_Zeilen : Word;
  Font_Bez, Zeichenbreite0 : Byte;

  KopieBWS : Boolean;

  // Variablen zum Bild
  BILD_Adresse : Adress;





{--------------------------------------------------------------}
{ functions }


// *** INITIALISIERUNG ***

procedure InitPorts;
begin
{
Grafik-Controller (GCON) (ATmega1284P-PU)
=================

D0..D7 = 8 (A0..A7)
A0..A16 = 17 (B0..B7, C0..C7, D2)


RX = 1 (D0)
TX = 1 (D1)

/WE = 1 (D3)
/OE = 1 (D4)

/VS1_GS2 = 1 (D5)
(0 = Counter (V) auf S1, Controller (G) auf S2)
(1 = Counter (V) auf S2, Controller (G) auf S1)

DIR = 1 (D6)

LED_GN = 1 (D7)

Summe = 32

}

  DDRA:=  %00000000;
  DDRB:=  %11111111;
  DDRC:=  %11111111;


// LED_GN = 1 (D7),/CE = 1 (D6),/VS1_GS2 = 0 (D5),/OE = 1 (D4),
// /WE = 1 (D3),A16  (D2),TX0 (D1),RX0 (D0)
  DDRD:=  %11111110;
  PORTD:= %01011000;

end InitPorts;




procedure INIT_SER_PORT;
begin

    UBRR0 := UBRR_VAL;

    UCSR0B := %00011000;

{
URSEL UMSEL UPM1 UPM0 USBS UCSZ1 UCSZ0 UCPOL
    1     0    0    0    0     1     1     0
}
    UCSR0C := %10000110;


end INIT_SER_PORT;





procedure Init_BILD_Adresse;
begin
  BILD_Adresse:=ArrToAdress(BILD5_256);
  ICON_Adress:=ArrToAdress(EBENE1_Array);
end Init_BILD_Adresse;




// *** Basis-Programme ***



PROCEDURE setze_Font(FontGroesse : Byte);
BEGIN
  Font_Bez:=FontGroesse;

  CASE FontGroesse OF

    fnt_6x8 : Zeichenbreite := 6;
              Zeichenbreite0 := 5;
              Zeichenhoehe  := 10; // 8 + 2 = 10

              X_ANF := 12;
              Y_ANF := 1;

              max_Spalten := 79;  // 0..79 = 80
              max_Spalten2 := 78;
              max_Zeilen  := 23;  // 0..23 = 24
              max_Zeilen2 := 22;

    |

    fnt_8x8 : Zeichenbreite := 8;
              Zeichenbreite0 := 7;
              Zeichenhoehe  := 10; // 8 + 2 = 10

              X_ANF := 1;
              Y_ANF := 1;

              max_Spalten := 63;  // 0..63 = 64
              max_Spalten2 := 62;
              max_Zeilen  := 23;  // 0..23 = 24
              max_Zeilen2 := 22;

    |

    ELSE


    ENDCASE;

END setze_Font;




PROCEDURE wait_ms(Zeit : Word);
VAR
  w1 : Word;
  w2 : Word;
BEGIN
  for w1:=1 to Zeit do
    for w2:=1 to 720 do // 45 * MHz
    endfor;
  endfor;
END wait_ms;







procedure clear_RAM;
VAR
  x1: Word;
  x2 : Word;
  Adresse : Word;

begin
  ClearBit(A16);
  ClearBit(DIR);

  DDRA:=  %11111111;
  PortA:= BCOLOR;

  SetBit(WE);

  Adresse:=0;

  for x1:=0 to 255 do

    for x2:=0 to 511 do
      PORTB:= Low(Adresse);
      PORTC:= High(Adresse);
      ClearBit(WE);
      Inc(Adresse);
      SetBit(WE);
    endfor;

    if Adresse=0 then
      SetBit(A16);
    endif;

  endfor;

  SetBit(WE);

  DDRA:= %00000000;
  SetBit(DIR);

  SetBit(OE);
  ClearBit(A16);

end clear_RAM;




procedure ScreenRoll;
begin

{$I ScreenRoll.inc}

end ScreenRoll;





procedure Umschalten_Speicherseite;
begin

  if BitHigh(VS1_GS2) then
    ClearBit(VS1_GS2);
  else
    SetBit(VS1_GS2);
  endif;

end Umschalten_Speicherseite;






procedure Reset_Form;
begin

  FCOLOR:= cl_Black;
  BCOLOR:= cl_White;
  X_KUR_POS:= 0;
  Y_KUR_POS:= 0;

  X_PIX_POS:= X_ANF;
  Y_PIX_POS:= Y_ANF;

  clear_RAM;
  if KopieBWS then
    Umschalten_Speicherseite;
    clear_RAM;
    Umschalten_Speicherseite;
  endif;

end Reset_Form;



// ***********************
// *** Grafik-Ausgabe  ***
// ***********************


// schreibt ein Bild an die Position: X_PIX_POS / Y_PIX_POS
procedure schreibe_Bild;
VAR
  x1: Word;
  x2 : Byte;
  x_bis : Word;
  y_bis : Byte;
  B_Wert : Byte;
  Adresse : Word;
  B_Pointer : Word;
  Offset : Word;

begin
  ClearBit(A16);
  ClearBit(DIR);

  DDRA:=  %11111111;

  SetBit(WE);

  if Y_PIX_POS>127 then
     // Berechnung der Adresse (1.Pixel)
     Adresse:=Y_PIX_POS-128;
     Adresse:=Adresse*512;
     Adresse:=Adresse+X_PIX_POS;
     SetBit(A16);
  else
     // Berechnung der Adresse (1.Pixel)
     Adresse:=Y_PIX_POS*512;
     Adresse:=Adresse+X_PIX_POS;
     ClearBit(A16);
  endif;

  B_Pointer:=0;

  // Laden und Berechnen der Bild-Breite

  B_Wert:=ReadAdrArray(BILD_Adresse,B_Pointer);
  Inc(B_Pointer);
  x_bis:=ByteToWord(B_Wert);
  x_bis:=x_bis*256;

  B_Wert:=ReadAdrArray(BILD_Adresse,B_Pointer);
  Inc(B_Pointer);
  x1:=ByteToWord(B_Wert);
  x_bis:=x_bis+x1;

  Offset:=512-x_bis;

  Dec(x_bis);

  // Laden und Berechnen der Bild-Hoehe
  y_bis:=ReadAdrArray(BILD_Adresse,B_Pointer);
  Inc(B_Pointer);
  Dec(y_bis);

  // Kopieren des Bildes
  for x2:=0 to y_bis do

    for x1:=0 to x_bis do

      B_Wert:=ReadAdrArray(BILD_Adresse,B_Pointer);

      if B_Wert<$FF then

        PortA:= B_Wert;
        PORTB:= Low(Adresse);
        PORTC:= High(Adresse);
        ClearBit(WE);
        Inc(Adresse);
        Inc(B_Pointer);
        SetBit(WE);

      else

        Inc(Adresse);
        Inc(B_Pointer);

      endif;

    endfor; { for x1:=0 to x_bis do }


    if Adresse>$FE00 then

      Adresse:=X_PIX_POS;
      SetBit(A16);

    else

      if Adresse=0000 then
        Adresse:=X_PIX_POS;
        SetBit(A16);
      else
        Adresse:=Adresse+Offset;
      endif;

    endif; { if Adresse>$FE00 then }

  endfor; { for x2:=0 to y_bis do }

  SetBit(WE);

  DDRA:= %00000000;
  SetBit(DIR);

  SetBit(OE);
  ClearBit(A16);

end schreibe_Bild;














// liest alle Pixel der angegebenen Zeichenposition
// vom VRAM und schreibt sie in das CHAR_Array
procedure VRAM_to_ICON_Array(X_PIXEL_POS, Y_PIXEL_POS : Word);
var
  Adresse : Word;
  x_cnt : Byte;
  y_cnt : Byte;
  Array_Pos : Word;
begin

  // Zeichen aus VRAM lesen -> Zeichen in Puffer schreiben
  SetBit(WE);

  DDRA:= %00000000;
  SetBit(DIR);

  SetBit(OE);
  ClearBit(A16);

  if Y_PIXEL_POS>127 then
     // Berechnung der Adresse (1.Pixel)
     Adresse:=Y_PIXEL_POS-128;
     Adresse:=Adresse*512;
     Adresse:=Adresse+X_PIXEL_POS;
     SetBit(A16);
  else
     // Berechnung der Adresse (1.Pixel)
     Adresse:=Y_PIXEL_POS*512;
     Adresse:=Adresse+X_PIXEL_POS;
     ClearBit(A16);
  endif;

  Array_Pos:=0;

  // Breite*Hoehe = 26x16=416
  for y_cnt:=0 to 15 do

    for x_cnt:=0 to 25 do

      PORTB:= Low(Adresse);
      PORTC:= High(Adresse);

      ClearBit(OE);
      INC(Adresse);

      WriteAdrArray(ICON_Adress,Array_Pos):=PINA;

      INC(Array_Pos);
      SetBit(OE);

    endfor; { for x_cnt:=0 to 25 do }


    if Adresse>$FE00 then

      Adresse:=X_PIXEL_POS;
      SetBit(A16);

    else

      if Adresse=0000 then
        Adresse:=X_PIXEL_POS;
        SetBit(A16);
      else
        Adresse:=Adresse+486; // 512-26=486
      endif;

    endif; { if Adresse>$FE00 then }

  endfor; { for y_cnt:=0 to 15 do }

  // VRAM Steuerleitungen auf Grundstellung
  SetBit(WE);
  DDRA:= %00000000;
  SetBit(DIR);
  SetBit(OE);
  ClearBit(A16);

end VRAM_to_ICON_Array;



procedure rette_ICON;
begin

  if BitHigh(VS1_GS2) then
    ICON_Adress:=ArrToAdress(EBENE1_Array);
    E1_X_PIX_POS:=X_PIX_POS;
    E1_Y_PIX_POS:=Y_PIX_POS;
    VRAM_to_ICON_Array(E1_X_PIX_POS,E1_Y_PIX_POS);
  else
    ICON_Adress:=ArrToAdress(EBENE2_Array);
    E2_X_PIX_POS:=X_PIX_POS;
    E2_Y_PIX_POS:=Y_PIX_POS;
    VRAM_to_ICON_Array(E2_X_PIX_POS,E2_Y_PIX_POS);
  endif; { if BitHigh(VS1_GS2) then }

end rette_ICON;













procedure ICON_Array_to_VRAM(X_PIXEL_POS, Y_PIXEL_POS : Word);
var
  Adresse : Word;
  x_cnt : Byte;
  y_cnt : Byte;
  Array_Pos : Word;
begin

  // ICON aus ARRAY lesen -> VRAM
    ClearBit(A16);
    ClearBit(DIR);

    DDRA:=  %11111111;

  if Y_PIXEL_POS>127 then
     // Berechnung der Adresse (1.Pixel)
     Adresse:=Y_PIXEL_POS-128;
     Adresse:=Adresse*512;
     Adresse:=Adresse+X_PIXEL_POS;
     SetBit(A16);
  else
     // Berechnung der Adresse (1.Pixel)
     Adresse:=Y_PIXEL_POS*512;
     Adresse:=Adresse+X_PIXEL_POS;
     ClearBit(A16);
  endif;

  Array_Pos:=0;

  // Breite*Hoehe = 26x16=416
  for y_cnt:=0 to 15 do

    for x_cnt:=0 to 25 do

      PORTA:= ReadAdrArray(ICON_Adress,Array_Pos);
      PORTB:= Low(Adresse);
      PORTC:= High(Adresse);

      ClearBit(WE);
      INC(Adresse);
      INC(Array_Pos);
      SetBit(WE);

    endfor; { for x_cnt:=0 to 25 do }

    if Adresse>$FE00 then

      Adresse:=X_PIXEL_POS;
      SetBit(A16);

    else

      if Adresse=0000 then
        Adresse:=X_PIXEL_POS;
        SetBit(A16);
      else
        Adresse:=Adresse+486; // 512-26=486
      endif;

    endif; { if Adresse>$FE00 then }

  endfor; { for y_cnt:=0 to 15 do }

  // VRAM Steuerleitungen auf Grundstellung
  SetBit(WE);
  DDRA:= %00000000;
  SetBit(DIR);
  SetBit(OE);
  ClearBit(A16);

end ICON_Array_to_VRAM;




procedure zurueck_ICON;
begin

  if BitHigh(VS1_GS2) then
    ICON_Adress:=ArrToAdress(EBENE1_Array);
    ICON_Array_to_VRAM(E1_X_PIX_POS,E1_Y_PIX_POS);
  else
    ICON_Adress:=ArrToAdress(EBENE2_Array);
    ICON_Array_to_VRAM(E2_X_PIX_POS,E2_Y_PIX_POS);
  endif; { if BitHigh(VS1_GS2) then }

end zurueck_ICON;















procedure kopiere_Kreis;
var
  K_Zaehler : Word;
begin

  BILD_Adresse:=ArrToAdress(KREIS2A);

  rette_ICON; // sichern der X und Y-Position und Sichern des Icons
  Umschalten_Speicherseite;
  rette_ICON;
  Umschalten_Speicherseite;

  Y_PIX_POS:= 20;
  for K_Zaehler:=10 to 250 do

    zurueck_ICON;
    X_PIX_POS:=K_Zaehler;
    rette_ICON; // sichern der X und Y-Position und Sichern des Icons
    schreibe_Bild;
    Umschalten_Speicherseite;

  endfor;

  zurueck_ICON;
  Umschalten_Speicherseite;
  zurueck_ICON;
  Umschalten_Speicherseite;

  X_KUR_POS:= 0;
  Y_KUR_POS:= 8;

  X_PIX_POS:= X_ANF;
  Y_PIX_POS:= Y_KUR_POS*10;
  Y_PIX_POS:= Y_PIX_POS+Y_ANF;

end kopiere_Kreis;







procedure kopiere_Hintergrundbild;
begin

 // Bild zeichnen
  X_PIX_POS:= 0;
  Y_PIX_POS:= 0;
  schreibe_Bild;

  X_PIX_POS:= 128;
  Y_PIX_POS:= 0;
  schreibe_Bild;

  X_PIX_POS:= 256;
  Y_PIX_POS:= 0;
  schreibe_Bild;

  X_PIX_POS:= 384;
  Y_PIX_POS:= 0;
  schreibe_Bild;

  if KopieBWS then

    Umschalten_Speicherseite;

    X_PIX_POS:= 0;
    Y_PIX_POS:= 0;

    schreibe_Bild;

    X_PIX_POS:= 128;
    Y_PIX_POS:= 0;
    schreibe_Bild;

    X_PIX_POS:= 256;
    Y_PIX_POS:= 0;
    schreibe_Bild;

    X_PIX_POS:= 384;
    Y_PIX_POS:= 0;
    schreibe_Bild;

    Umschalten_Speicherseite;

  endif; { if KopieBWS then }


  X_KUR_POS:= 0;
  Y_KUR_POS:= 14;

  X_PIX_POS:= X_ANF;
  Y_PIX_POS:= Y_KUR_POS*10;
  Y_PIX_POS:= Y_PIX_POS+Y_ANF;

end kopiere_Hintergrundbild;













procedure SET_POINT;
VAR
  Adresse : Word;

begin

  if Y_PIX_POS>127 then
     // Berechnung der Adresse (1.Pixel)
     Adresse:=Y_PIX_POS-128;
     Adresse:=Adresse*512;
     Adresse:=Adresse+X_PIX_POS;
     SetBit(A16);
  else
     // Berechnung der Adresse (1.Pixel)
     Adresse:=Y_PIX_POS*512;
     Adresse:=Adresse+X_PIX_POS;
     ClearBit(A16);
  endif;

  ClearBit(DIR);

  DDRA:=  %11111111;

  SetBit(WE);

  PortA:=PIXCOLOR;

  PORTB:= Low(Adresse);
  PORTC:= High(Adresse);
  ClearBit(WE);
  NOP;
  NOP;
  NOP;
  NOP;
  SetBit(WE);

  DDRA:= %00000000;
  SetBit(DIR);

  SetBit(OE);
  ClearBit(A16);

end SET_POINT;



procedure GrafikDrawPoint;
begin

  SET_POINT;
  if KopieBWS then
    Umschalten_Speicherseite;
    SET_POINT;
    Umschalten_Speicherseite;
  endif;

end GrafikDrawPoint;






procedure SET_POINT_POS(X_POINT_POS, Y_POINT_POS : Word);
VAR
  Adresse : Word;

begin

  if Y_POINT_POS>127 then
     // Berechnung der Adresse (1.Pixel)
     Adresse:=Y_POINT_POS-128;
     Adresse:=Adresse*512;
     Adresse:=Adresse+X_POINT_POS;
     SetBit(A16);
  else
     // Berechnung der Adresse (1.Pixel)
     Adresse:=Y_POINT_POS*512;
     Adresse:=Adresse+X_POINT_POS;
     ClearBit(A16);
  endif;

  ClearBit(DIR);

  DDRA:=  %11111111;

  SetBit(WE);

  PortA:=PIXCOLOR;

  PORTB:= Low(Adresse);
  PORTC:= High(Adresse);
  ClearBit(WE);
  NOP;
  NOP;
  NOP;
  NOP;
  SetBit(WE);

  DDRA:= %00000000;
  SetBit(DIR);

  SetBit(OE);
  ClearBit(A16);

end SET_POINT_POS;





// zeichnet eine Linie zwischen den Punkten:
// P1(X_PIX_POS,Y_PIX_POS) und P2(X_END_POS,Y_END_POS)
procedure LINE_TO;
var
  x1 : Word;
  xp_anf : Word;
  xp_end : Word;
  y1 : Word;
  yp_anf : Word;
  yp_end : Word;

begin

  // da ohne Vorzeichen gearbeitet wird muss
  // als Erstes festgestellt werden, welcher
  // X-Wert groesser ist ...
  if X_PIX_POS<X_END_POS then
    delta_x:=X_END_POS-X_PIX_POS;
    xp_anf:=X_PIX_POS;
    xp_end:=X_END_POS;
  else
    delta_x:=X_PIX_POS-X_END_POS;
    xp_anf:=X_END_POS;
    xp_end:=X_PIX_POS;
  endif;

  // ... und welcher Y-Wert groesser ist
  if Y_PIX_POS<Y_END_POS then
    delta_y:=Y_END_POS-Y_PIX_POS;
    yp_anf:=Y_PIX_POS;
    yp_end:=Y_END_POS;
  else
    delta_y:=Y_PIX_POS-Y_END_POS;
    yp_anf:=Y_END_POS;
    yp_end:=Y_PIX_POS;
  endif;

  // die Schleife laeuft am Besten entlang der
  // laengeren Seite
  if delta_x>delta_y then

    // Ausnahmebehandlung: Y_PIX_POS=Y_END_POS (waagerechter Strich)
    if delta_y=0 then
      // Durchlaufen aller X-Werte
      for x1:=xp_anf to xp_end do
        SET_POINT_POS(x1,Y_PIX_POS);
      endfor;
    else

      if delta_y>127 then
        SHR(delta_x);
        SHR(delta_y);
      endif;

      if delta_x>0 then

         // Durchlaufen aller x-Werte
        for x1:=xp_anf to xp_end do

          // aktueller x-Wert mal Anstiegsfaktor ergibt den
          // aktuellen y-Wert
          y1:=x1*delta_y;
          y1:=y1 DIV delta_x;
          y1:=y_anf+y1;

          SET_POINT_POS(x1,y1);
        endfor;

      endif; { if delta_x>0 then }

    endif; { if delta_y=0 then }
  else

    // Ausnahmebehandlung: X_PIX_POS=X_END_POS (senkrechter Strich)
    if delta_x=0 then

      // Durchlaufen aller y-Werte
      for y1:=yp_anf to yp_end do
        SET_POINT_POS(X_PIX_POS,y1);
       endfor;

    else

      if delta_x>127 then
        SHR(delta_x);
        SHR(delta_y);
      endif;

      if delta_y>0 then

        // Durchlaufen aller y-Werte
        for y1:=yp_anf to yp_end do

          // aktueller y-Wert mal Anstiegsfaktor ergibt den
          // aktuellen x-Wert
          x1:=y1*delta_x;
          x1:=x1 DIV delta_y;
          x1:=x_anf+x1;

          SET_POINT_POS(x1,y1);
        endfor;

      endif; { if delta_y>0 then }

    endif; { if delta_x=0 then }

  endif; { if delta_x>delta_y then }

end LINE_TO;





// Zeichnen einer Linie
procedure GrafikDrawLine;
begin

  LINE_TO;
  if KopieBWS then
    Umschalten_Speicherseite;
    LINE_TO;
    Umschalten_Speicherseite;
  endif;

  X_PIX_POS:=X_END_POS;
  Y_PIX_POS:=Y_END_POS;

end GrafikDrawLine;










procedure DRAW_REC;
VAR
  Adresse : Word;
  w1 : Word;
  YR_POS : Word;
begin

  if Y_PIX_POS>127 then
     // Berechnung der Adresse (1.Pixel)
     YR_POS:=Y_PIX_POS-128;
     Adresse:=YR_POS;
     Adresse:=Adresse*512;
     Adresse:=Adresse+X_PIX_POS;
     SetBit(A16);
  else
     // Berechnung der Adresse (1.Pixel)
     YR_POS:=Y_PIX_POS;
     Adresse:=Y_PIX_POS*512;
     Adresse:=Adresse+X_PIX_POS;
     ClearBit(A16);
  endif;

  ClearBit(DIR);

  DDRA:=  %11111111;

  SetBit(WE);

  PortA:=PIXCOLOR;






  // oberer waagerechter Strich
  for w1:=X_PIX_POS to X_END_POS do

    PORTB:= Low(Adresse);
    PORTC:= High(Adresse);
    ClearBit(WE);
    INC(Adresse);
    SetBit(WE);

  endfor;

  DEC(Adresse);

  // rechter senkrechter Strich
  for w1:=Y_PIX_POS to Y_END_POS do

    Adresse:=Adresse+512;

    INC(YR_POS);
    if YR_POS=128 then
      YR_POS:=0;
      SetBit(A16);
      Adresse:=X_END_POS;
    endif;

    PORTB:= Low(Adresse);
    PORTC:= High(Adresse);
    ClearBit(WE);
    NOP;
    NOP;
    NOP;
    NOP;
    SetBit(WE);

  endfor;




  if Y_PIX_POS>127 then
     // Berechnung der Adresse (1.Pixel)
     YR_POS:=Y_PIX_POS-128;
     Adresse:=YR_POS;
     Adresse:=Adresse*512;
     Adresse:=Adresse+X_PIX_POS;
     SetBit(A16);
  else
     // Berechnung der Adresse (1.Pixel)
     YR_POS:=Y_PIX_POS;
     Adresse:=Y_PIX_POS*512;
     Adresse:=Adresse+X_PIX_POS;
     ClearBit(A16);
  endif;




  // linker senkrechter Strich
  for w1:=Y_PIX_POS to Y_END_POS do

    Adresse:=Adresse+512;

    INC(YR_POS);
    if YR_POS=128 then
      YR_POS:=0;
      SetBit(A16);
      Adresse:=X_PIX_POS;
    endif;

    PORTB:= Low(Adresse);
    PORTC:= High(Adresse);
    ClearBit(WE);
    NOP;
    NOP;
    NOP;
    NOP;
    SetBit(WE);

  endfor;



  // unterer waagerechter Strich
  for w1:=X_PIX_POS to X_END_POS do

    PORTB:= Low(Adresse);
    PORTC:= High(Adresse);
    ClearBit(WE);
    INC(Adresse);
    SetBit(WE);

  endfor;



  DDRA:= %00000000;
  SetBit(DIR);

  SetBit(OE);
  ClearBit(A16);

end DRAW_REC;







procedure GrafikDrawRectangle;
begin

  DRAW_REC;

  if KopieBWS then
    Umschalten_Speicherseite;
    DRAW_REC;
    Umschalten_Speicherseite;
  endif;

  X_PIX_POS:=X_END_POS;
  Y_PIX_POS:=Y_END_POS;

end GrafikDrawRectangle;







procedure DRAW_FULL_REC;
VAR
  Adresse : Word;
  w1 : Word;
  w2 : Word;
begin

  ClearBit(DIR);

  DDRA:=  %11111111;

  SetBit(WE);

  PortA:=PIXCOLOR;

  for w2:=Y_PIX_POS to Y_END_POS do

    if w2>127 then
       // Berechnung der Adresse (1.Pixel)
       Adresse:=w2-128;
       Adresse:=Adresse*512;
       Adresse:=Adresse+X_PIX_POS;
       SetBit(A16);
    else
       // Berechnung der Adresse (1.Pixel)
       Adresse:=w2*512;
       Adresse:=Adresse+X_PIX_POS;
       ClearBit(A16);
    endif;

    // waagerechter Strich
    for w1:=X_PIX_POS to X_END_POS do

      PORTB:= Low(Adresse);
      PORTC:= High(Adresse);
      ClearBit(WE);
      INC(Adresse);
      SetBit(WE);

    endfor;

  endfor;

  DDRA:= %00000000;
  SetBit(DIR);

  SetBit(OE);
  ClearBit(A16);

end DRAW_FULL_REC;





procedure GrafikDrawFullRectangle;
begin

  DRAW_FULL_REC;

  if KopieBWS then
    Umschalten_Speicherseite;
    DRAW_FULL_REC;
    Umschalten_Speicherseite;
  endif;

  X_PIX_POS:=X_END_POS;
  Y_PIX_POS:=Y_END_POS;

end GrafikDrawFullRectangle;





function Sinus(Winkel, Radius : Word) : Word;
var
  w1 : Word;
begin

  w1:=SinTab[Winkel];
  Sin_dw1:=WordTODWord(w1);
  Sin_dw2:=WordTODWord(Radius);
  Sin_dw1:=Sin_dw1*Sin_dw2;
  Sin_dw1:=Sin_dw1 DIV $FFFF;
  w1:=DWLow(Sin_dw1);

  Result:=w1;
end Sinus;







// Seitenverhaeltnis Breite zu Hoehe -> 170 / 105 = 34 / 21

procedure GrafikZeichneKreis_on_Rampage(Radius : Word);
var
  x1 : Word;
  y1 : Word;
  Winkel : Word;
  x_Radius : Word;
begin
  x_Radius:=Radius*34;
  x_Radius:=x_Radius DIV 21;

  // Kreissegment 0..89 Grad
  for Winkel:=0 to 199 do
    // Delta Radius fuer x1
    x1:=Sinus(Winkel,x_Radius);

    // Delta Radius fuer y1
    y1:=200-Winkel;
    y1:=Sinus(y1,Radius);

    // x1 fuer Kreis-Punkt
    x1:=X_PIX_POS+x1;

    // y1 fuer Kreis-Punkt
    y1:=Y_PIX_POS-y1;

    SET_POINT_POS(x1,y1);
  endfor;

  // Kreissegment 90..179 Grad
  for Winkel:=0 to 199 do
    // Delta Radius fuer x1
    x1:=200-Winkel;
    x1:=Sinus(x1,x_Radius);

    // Delta Radius fuer y1
    y1:=Sinus(Winkel,Radius);

    // x1 fuer Kreis-Punkt
    x1:=X_PIX_POS+x1;

    // y1 fuer Kreis-Punkt
    y1:=Y_PIX_POS+y1;

    SET_POINT_POS(x1,y1);
  endfor;

  // Kreissegment 180..269 Grad
  for Winkel:=0 to 199 do
    // Delta Radius fuer x1
    x1:=Sinus(Winkel,x_Radius);

    // Delta Radius fuer y1
    y1:=200-Winkel;
    y1:=Sinus(y1,Radius);

    // x1 fuer Kreis-Punkt
    x1:=X_PIX_POS-x1;

    // y1 fuer Kreis-Punkt
    y1:=Y_PIX_POS+y1;

    SET_POINT_POS(x1,y1);

  endfor;

  // Kreissegment 270..359 Grad
  for Winkel:=0 to 199 do
    // Delta Radius fuer x1
    x1:=200-Winkel;
    x1:=Sinus(x1,x_Radius);

    // Delta Radius fuer y1
    y1:=Sinus(Winkel,Radius);

    // x1 fuer Kreis-Punkt
    x1:=X_PIX_POS-x1;

    // y1 fuer Kreis-Punkt
    y1:=Y_PIX_POS-y1;

    SET_POINT_POS(x1,y1);
  endfor;

end GrafikZeichneKreis_on_Rampage;









procedure GrafikZeichneKreis(Radius : Word);
begin

  GrafikZeichneKreis_on_Rampage(Radius);

  if KopieBWS then
    Umschalten_Speicherseite;
    GrafikZeichneKreis_on_Rampage(Radius);
    Umschalten_Speicherseite;
  endif;

end GrafikZeichneKreis;







// von X_PIX_POS nach X_END_POS
// an der Stelle Y_PIX_POS
procedure Zeichne_waagerechten_Strich;
VAR
  Adresse : Word;
  w1 : Word;
begin

  ClearBit(DIR);

  DDRA:=  %11111111;

  SetBit(WE);

  PortA:=PIXCOLOR;

  if Y_PIX_POS>127 then
     // Berechnung der Adresse (1.Pixel)
     Adresse:=Y_PIX_POS-128;
     Adresse:=Adresse*512;
     Adresse:=Adresse+X_PIX_POS;
     SetBit(A16);
  else
     // Berechnung der Adresse (1.Pixel)
     Adresse:=Y_PIX_POS*512;
     Adresse:=Adresse+X_PIX_POS;
     ClearBit(A16);
  endif;

  // waagerechter Strich
  for w1:=X_PIX_POS to X_END_POS do

    PORTB:= Low(Adresse);
    PORTC:= High(Adresse);
    ClearBit(WE);
    INC(Adresse);
    SetBit(WE);

  endfor;


  DDRA:= %00000000;
  SetBit(DIR);

  SetBit(OE);
  ClearBit(A16);

end Zeichne_waagerechten_Strich;








procedure GrafikZeichneVollenKreis_on_Rampage(Radius : Word);
var
  X_PIX_POS2 : Word;
  Y_PIX_POS2 : Word;
  Winkel : Word;
  x_Radius : Word;
begin
  x_Radius:=Radius*34;
  x_Radius:=x_Radius DIV 21;

  X_PIX_POS2:=X_PIX_POS;
  Y_PIX_POS2:=Y_PIX_POS;

  // Kreissegment 0..89 Grad
  for Winkel:=0 to 199 do

    // XXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
    // Delta Radius fuer X-Position
    X_PIX_POS:=Sinus(Winkel,x_Radius);

   // X_END_POS ist rechte X-Position
    X_END_POS:=X_PIX_POS2+X_PIX_POS;

    // X_PIX_POS ist linke X-Position
    X_PIX_POS:=X_PIX_POS2-X_PIX_POS;

    // YYYYYYYYYYYYYYYYYYYYYYYYYYYYYY
    // Delta Radius fuer y1
    Y_PIX_POS:=200-Winkel;
    Y_PIX_POS:=Sinus(Y_PIX_POS,Radius);

    // y1 fuer Kreis-Punkt
    Y_PIX_POS:=Y_PIX_POS2-Y_PIX_POS;

    Zeichne_waagerechten_Strich;
  endfor;

  // Kreissegment 90..179 Grad
  for Winkel:=0 to 199 do
    // XXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
    // Delta Radius fuer X-Position
    X_PIX_POS:=200-Winkel;
    X_PIX_POS:=Sinus(X_PIX_POS,x_Radius);

   // X_END_POS ist rechte X-Position
    X_END_POS:=X_PIX_POS2+X_PIX_POS;

    // X_PIX_POS ist linke X-Position
    X_PIX_POS:=X_PIX_POS2-X_PIX_POS;

    // YYYYYYYYYYYYYYYYYYYYYYYYYYYYYY
    // Delta Radius fuer Y-Position
    Y_PIX_POS:=Sinus(Winkel,Radius);

    // y1 fuer Kreis-Punkt
    Y_PIX_POS:=Y_PIX_POS2+Y_PIX_POS;

    Zeichne_waagerechten_Strich;
  endfor;

  X_PIX_POS:=X_PIX_POS2;
  Y_PIX_POS:=Y_PIX_POS2;

end GrafikZeichneVollenKreis_on_Rampage;








procedure GrafikZeichneVollenKreis(Radius : Word);
begin

  GrafikZeichneVollenKreis_on_Rampage(Radius);

  if KopieBWS then
    Umschalten_Speicherseite;
    GrafikZeichneVollenKreis_on_Rampage(Radius);
    Umschalten_Speicherseite;
  endif;

end GrafikZeichneVollenKreis;



















// ***********************
// *** Zeichen-Ausgabe ***
// ***********************




// schreibt ein Zeichen in die aktuelle RamPage
// ohne Kursorbewegung
procedure Write_Char_6x8;
var
  FontPtr : Word;
  x_cnt : Byte;
  y_cnt : Word;
  FontColor : Byte;
  FontZeile : Byte;
  Adresse : Word;
begin

  ClearBit(DIR);
  DDRA:= %11111111;

  SetBit(WE);


  // Zeiger auf Zeichen in der Zeichentabelle
  FontPtr:=ByteToWord(Zeichen);
  FontPtr:=FontPtr-$20;
  FontPtr:=FontPtr*8;


  // Sonderfall: Y_KUR_POS=12
  if Y_KUR_POS=12 then

    // Berechnung der Adresse (1.Pixel)
    Adresse:=Y_PIX_POS*512;
    Adresse:=Adresse+X_PIX_POS;
    ClearBit(A16);

    // zuerst nur Y_PIX_POS 121...127
    for y_cnt:=0 to 6 do

      FontZeile:=font6x8[FontPtr];
      INC(FontPtr);
      SHR(FontZeile);
      SHR(FontZeile);

      for x_cnt:=0 to Zeichenbreite0 do

        FontColor:=FontZeile and %00000001;
        if FontColor = 0 then
          PORTA:= BCOLOR;
        else
          PORTA:= FCOLOR;
        endif;

        PORTB:= Low(Adresse);
        PORTC:= High(Adresse);

        ClearBit(WE);
        SHR(FontZeile);
        INC(Adresse);
        SetBit(WE);

      endfor; { for x_cnt:=1 to 6 do }
    
      Adresse:=Adresse+506;  // Adresse + PixelProZeile - Zeichenbreite

    endfor; { for y_cnt:=0 to 6 do }


    // ... dann Y_PIX_POS 128

    // Berechnung der Adresse (1.Pixel)
    Adresse:=X_PIX_POS;
    SetBit(A16);

    FontZeile:=font6x8[FontPtr];
    SHR(FontZeile);
    SHR(FontZeile);

    for x_cnt:=0 to Zeichenbreite0 do

      FontColor:=FontZeile and %00000001;
      if FontColor = 0 then
        PORTA:= BCOLOR;
      else
        PORTA:= FCOLOR;
      endif;

      PORTB:= Low(Adresse);
      PORTC:= High(Adresse);

      ClearBit(WE);
      SHR(FontZeile);
      INC(Adresse);
      SetBit(WE);

    endfor; { for y_cnt:=0 to 5 do }


  // Y_KUR_POS<>12 ... der Normalfall
  else
  
    if Y_KUR_POS>12 then
      // Berechnung der Adresse (1.Pixel)
      Adresse:=Y_PIX_POS-128;
      Adresse:=Adresse*512;
      Adresse:=Adresse+X_PIX_POS;
      SetBit(A16);
    else
      // Berechnung der Adresse (1.Pixel)
      Adresse:=Y_PIX_POS*512;
      Adresse:=Adresse+X_PIX_POS;
      ClearBit(A16);
    endif;

    for y_cnt:=0 to 7 do

      FontZeile:=font6x8[FontPtr];
      INC(FontPtr);
      SHR(FontZeile);
      SHR(FontZeile);

      for x_cnt:=0 to Zeichenbreite0 do

        FontColor:=FontZeile and %00000001;
        if FontColor = 0 then
          PORTA:= BCOLOR;
        else
          PORTA:= FCOLOR;
        endif;

        PORTB:= Low(Adresse);
        PORTC:= High(Adresse);

        ClearBit(WE);
        SHR(FontZeile);
        INC(Adresse);
        SetBit(WE);

      endfor; { for x_cnt:=0 to 5 do }
    
      Adresse:=Adresse+506; // Adresse + PixelProZeile - Zeichenbreite

    endfor; { for y_cnt:=0 to 7 do }

  endif; { if Y_KUR_POS=12 then }

  // Grundeinstellungen
  SetBit(WE);
  DDRA:= %00000000;
  SetBit(DIR);
  SetBit(OE);
  ClearBit(A16);

end Write_Char_6x8;







// schreibt ein Zeichen in die aktuelle RamPage
// ohne Kursorbewegung
procedure Write_Char_8x8;
var
  FontPtr : Word;
  x_cnt : Byte;
  y_cnt : Word;
  FontColor : Byte;
  FontZeile : Byte;
  Adresse : Word;
begin

  ClearBit(DIR);
  DDRA:= %11111111;

  SetBit(WE);


  // Zeiger auf Zeichen in der Zeichentabelle
  FontPtr:=ByteToWord(Zeichen);
  FontPtr:=FontPtr-$20;
  FontPtr:=FontPtr*8;


  // Sonderfall: Y_KUR_POS=12
  if Y_KUR_POS=12 then

    // Berechnung der Adresse (1.Pixel)
    Adresse:=Y_PIX_POS*512;
    Adresse:=Adresse+X_PIX_POS;
    ClearBit(A16);

    // zuerst nur Y_PIX_POS 121...127
    for y_cnt:=0 to 6 do

      FontZeile:=font8x8[FontPtr];
      INC(FontPtr);

      for x_cnt:=0 to Zeichenbreite0 do

        FontColor:=FontZeile and %00000001;
        if FontColor = 0 then
          PORTA:= BCOLOR;
        else
          PORTA:= FCOLOR;
        endif;

        PORTB:= Low(Adresse);
        PORTC:= High(Adresse);

        ClearBit(WE);
        SHR(FontZeile);
        INC(Adresse);
        SetBit(WE);

      endfor; { for x_cnt:=1 to 6 do }

      Adresse:=Adresse+504; // Adresse + PixelProZeile - Zeichenbreite

    endfor; { for y_cnt:=0 to 6 do }


    // ... dann Y_PIX_POS 128

    // Berechnung der Adresse (1.Pixel)
    Adresse:=X_PIX_POS;
    SetBit(A16);

    FontZeile:=font8x8[FontPtr];

    for x_cnt:=0 to Zeichenbreite0 do

      FontColor:=FontZeile and %00000001;
      if FontColor = 0 then
        PORTA:= BCOLOR;
      else
        PORTA:= FCOLOR;
      endif;

      PORTB:= Low(Adresse);
      PORTC:= High(Adresse);

      ClearBit(WE);
      SHR(FontZeile);
      INC(Adresse);
      SetBit(WE);

    endfor; { for y_cnt:=0 to 5 do }


  // Y_KUR_POS<>12 ... der Normalfall
  else

    if Y_KUR_POS>12 then
      // Berechnung der Adresse (1.Pixel)
      Adresse:=Y_PIX_POS-128;
      Adresse:=Adresse*512;
      Adresse:=Adresse+X_PIX_POS;
      SetBit(A16);
    else
      // Berechnung der Adresse (1.Pixel)
      Adresse:=Y_PIX_POS*512;
      Adresse:=Adresse+X_PIX_POS;
      ClearBit(A16);
    endif;

    for y_cnt:=0 to 7 do

      FontZeile:=font8x8[FontPtr];
      INC(FontPtr);

      for x_cnt:=0 to Zeichenbreite0 do

        FontColor:=FontZeile and %00000001;
        if FontColor = 0 then
          PORTA:= BCOLOR;
        else
          PORTA:= FCOLOR;
        endif;

        PORTB:= Low(Adresse);
        PORTC:= High(Adresse);

        ClearBit(WE);
        SHR(FontZeile);
        INC(Adresse);
        SetBit(WE);

      endfor; { for x_cnt:=0 to 5 do }

      Adresse:=Adresse+504; // Adresse + PixelProZeile(512) - Zeichenbreite

    endfor; { for y_cnt:=0 to 7 do }

  endif; { if Y_KUR_POS=12 then }

  // Grundeinstellungen
  SetBit(WE);
  DDRA:= %00000000;
  SetBit(DIR);
  SetBit(OE);
  ClearBit(A16);

end Write_Char_8x8;






procedure Write_Char;
begin

  CASE Font_Bez OF
    fnt_6x8 : Write_Char_6x8;
    |
    fnt_8x8 : Write_Char_8x8;
    |
  ENDCASE;

end Write_Char;







procedure VRAM_to_Kursor_Array;
var
  Adresse : Word;
  x_cnt : Byte;
  y_cnt : Byte;
  Array_Pos : Word;
begin

  // Zeichen aus VRAM lesen -> Zeichen in Puffer schreiben
  SetBit(WE);

  DDRA:= %00000000;
  SetBit(DIR);

  SetBit(OE);
  ClearBit(A16);


  // Sonderfall: Y_KUR_POS=12
  if Y_KUR_POS=12 then

    // Berechnung der Adresse (1.Pixel)
    Adresse:=Y_PIX_POS*512;
    Adresse:=Adresse+X_PIX_POS;
    ClearBit(A16);
    Array_Pos:=0;

    for y_cnt:=0 to 6 do
 
      for x_cnt:=0 to Zeichenbreite0 do
 
        PORTB:= Low(Adresse);
        PORTC:= High(Adresse);

        ClearBit(OE);
        INC(Adresse);
        Kursor_Array[Array_Pos]:=PINA;
        INC(Array_Pos);
        SetBit(OE);

      endfor; { for x_cnt:=0 to 5 do } 
    
      Adresse:=Adresse+512;
      Adresse:=Adresse-Zeichenbreite;

    endfor; { for y_cnt:=0 to 6 do } 


    // ... dann Y_PIX_POS 128

    // Berechnung der Adresse (1.Pixel)
    Adresse:=X_PIX_POS;
    SetBit(A16);

    for x_cnt:=0 to 5 do

      PORTB:= Low(Adresse);
      PORTC:= High(Adresse);

      ClearBit(OE);
      INC(Adresse);
      Kursor_Array[Array_Pos]:=PINA;
      INC(Array_Pos);
      SetBit(OE);

    endfor; { for y_cnt:=0 to 5 do }

  else

    if Y_KUR_POS>12 then
      // Berechnung der Adresse (1.Pixel)
      Adresse:=Y_PIX_POS-128;
      Adresse:=Adresse*512;
      Adresse:=Adresse+X_PIX_POS;
      SetBit(A16);
    else
      // Berechnung der Adresse (1.Pixel)
      Adresse:=Y_PIX_POS*512;
      Adresse:=Adresse+X_PIX_POS;
      ClearBit(A16);
    endif;

    Array_Pos:=0;

    for y_cnt:=0 to 7 do
 
      for x_cnt:=0 to Zeichenbreite0 do
 
        PORTB:= Low(Adresse);
        PORTC:= High(Adresse);

        NOP;
        NOP;

        ClearBit(OE);
        INC(Adresse);
        Kursor_Array[Array_Pos]:=PINA;
        INC(Array_Pos);
        SetBit(OE);

      endfor; { for x_cnt:=0 to 5 do } 
    
      Adresse:=Adresse+512;
      Adresse:=Adresse-Zeichenbreite;

    endfor; { for y_cnt:=0 to 7 do } 

  endif; { if Y_KUR_POS=12 then }

  // VRAM Steuerleitungen auf Grundstellung
  SetBit(WE);
  DDRA:= %00000000;
  SetBit(DIR);
  SetBit(OE);
  ClearBit(A16);

  KurArr_Aktuell:=true;

end VRAM_to_Kursor_Array;






procedure Kursor_Array_to_VRAM;
var
  Adresse : Word;
  x_cnt : Byte;
  y_cnt : Byte;
  Array_Pos : Word;
begin

  if KurArr_Aktuell then

    // Zeichen aus Puffer lesen -> in VRAM schreiben
    ClearBit(A16);
    ClearBit(DIR);

    DDRA:=  %11111111;

    // Sonderfall: Y_KUR_POS=12
    if Y_KUR_POS=12 then

      // Berechnung der Adresse (1.Pixel)
      Adresse:=Y_PIX_POS*512;
      Adresse:=Adresse+X_PIX_POS;
      ClearBit(A16);
      Array_Pos:=0;

      for y_cnt:=0 to 6 do
 
        for x_cnt:=0 to Zeichenbreite0 do
 
          PORTA:= Kursor_Array[Array_Pos];
          PORTB:= Low(Adresse);
          PORTC:= High(Adresse);

          ClearBit(WE);
          INC(Adresse);
          INC(Array_Pos);
          SetBit(WE);

        endfor; 
    
        Adresse:=Adresse+512;
        Adresse:=Adresse-Zeichenbreite;

      endfor; 

      // ... dann Y_PIX_POS 128
 
      // Berechnung der Adresse (1.Pixel)
      Adresse:=X_PIX_POS;
      SetBit(A16);

      for x_cnt:=0 to Zeichenbreite0 do
  
        PORTA:= Kursor_Array[Array_Pos];
        PORTB:= Low(Adresse);
        PORTC:= High(Adresse);

        ClearBit(WE);
        INC(Adresse);
        INC(Array_Pos);
        SetBit(WE);

      endfor; { for x_cnt:=0 to 5 do }

    else

      if Y_KUR_POS>12 then
        // Berechnung der Adresse (1.Pixel)
        Adresse:=Y_PIX_POS-128;
        Adresse:=Adresse*512;
        Adresse:=Adresse+X_PIX_POS;
        SetBit(A16);
      else
        // Berechnung der Adresse (1.Pixel)
        Adresse:=Y_PIX_POS*512;
        Adresse:=Adresse+X_PIX_POS;
        ClearBit(A16);
      endif; { if Y_KUR_POS>12 then }

      Array_Pos:=0;

      for y_cnt:=0 to 7 do
 
        for x_cnt:=0 to Zeichenbreite0 do
 
          PORTA:= Kursor_Array[Array_Pos];
          PORTB:= Low(Adresse);
          PORTC:= High(Adresse);

          ClearBit(WE);
          INC(Adresse);
          INC(Array_Pos);
          SetBit(WE);

        endfor; { for x_cnt:=0 to 5 do }
    
        Adresse:=Adresse+512;
        Adresse:=Adresse-Zeichenbreite;

      endfor; { for y_cnt:=0 to 7 do }

    endif; { else Y_KUR_POS=12 then }

    // Grundeinstellungen
    SetBit(WE);
    DDRA:= %00000000;
    SetBit(DIR);
    SetBit(OE);
    ClearBit(A16);

  endif; { if KurArr_Aktuell then }

end Kursor_Array_to_VRAM;








procedure Schalte_Kursor_Ein;
begin

  VRAM_to_Kursor_Array;

  // Kursor schreiben
  Zeichen:=$5F;

  Write_Char;
  if KopieBWS then
    Umschalten_Speicherseite;
    Write_Char;
    Umschalten_Speicherseite;
  endif;

end Schalte_Kursor_Ein;





procedure Schalte_Kursor_Aus;
begin

  Kursor_Array_to_VRAM;

  if KopieBWS then
    Umschalten_Speicherseite;
    Kursor_Array_to_VRAM;
    Umschalten_Speicherseite;
  endif;

end Schalte_Kursor_Aus;













// liest alle Pixel der angegebenen Zeichenposition
// vom VRAM und schreibt sie in das CHAR_Array
procedure VRAM_to_CHAR_Array(X_PIXEL_POS, Y_PIXEL_POS : Word);
var
  Adresse : Word;
  x_cnt : Byte;
  y_cnt : Byte;
  Array_Pos : Word;
begin

  // Zeichen aus VRAM lesen -> Zeichen in Puffer schreiben
  SetBit(WE);

  DDRA:= %00000000;
  SetBit(DIR);

  SetBit(OE);
  ClearBit(A16);


  // Sonderfall: Y_PIXEL_POS=121
  if Y_PIXEL_POS=121 then

    // Berechnung der Adresse (1.Pixel)
    Adresse:=Y_PIXEL_POS*512;
    Adresse:=Adresse+X_PIXEL_POS;
    ClearBit(A16);
    Array_Pos:=0;

    // 121 bis 127
    for y_cnt:=0 to 6 do

      for x_cnt:=0 to Zeichenbreite0 do

        PORTB:= Low(Adresse);
        PORTC:= High(Adresse);

        ClearBit(OE);
        INC(Adresse);
        CHAR_Array[Array_Pos]:=PINA;
        INC(Array_Pos);
        SetBit(OE);

      endfor; { for x_cnt:=0 to 5 do }

      Adresse:=Adresse+512;
      Adresse:=Adresse-Zeichenbreite;

    endfor; { for y_cnt:=0 to 6 do }


    // ... dann Y_PIX_POS 128

    // Berechnung der Adresse (1.Pixel)
    Adresse:=X_PIXEL_POS;
    SetBit(A16);

    for x_cnt:=0 to Zeichenbreite0 do

      PORTB:= Low(Adresse);
      PORTC:= High(Adresse);

      ClearBit(OE);
      INC(Adresse);
      CHAR_Array[Array_Pos]:=PINA;
      INC(Array_Pos);
      SetBit(OE);

    endfor; { for y_cnt:=0 to 5 do }

  else

    if Y_PIXEL_POS>121 then
      // Berechnung der Adresse (1.Pixel)
      Adresse:=Y_PIXEL_POS-128;
      Adresse:=Adresse*512;
      Adresse:=Adresse+X_PIXEL_POS;
      SetBit(A16);
    else
      // Berechnung der Adresse (1.Pixel)
      Adresse:=Y_PIXEL_POS*512;
      Adresse:=Adresse+X_PIXEL_POS;
      ClearBit(A16);
    endif;

    Array_Pos:=0;

    for y_cnt:=0 to 7 do

      for x_cnt:=0 to Zeichenbreite0 do

        PORTB:= Low(Adresse);
        PORTC:= High(Adresse);

        NOP;
        NOP;

        ClearBit(OE);
        INC(Adresse);
        CHAR_Array[Array_Pos]:=PINA;
        INC(Array_Pos);
        SetBit(OE);

      endfor; { for x_cnt:=0 to 5 do }

      Adresse:=Adresse+512;
      Adresse:=Adresse-Zeichenbreite;

    endfor; { for y_cnt:=0 to 7 do }

  endif; { if Y_KURSOR=12 then }

  // VRAM Steuerleitungen auf Grundstellung
  SetBit(WE);
  DDRA:= %00000000;
  SetBit(DIR);
  SetBit(OE);
  ClearBit(A16);


end VRAM_to_CHAR_Array;

















// schreibt alle Pixel der angegebenen Zeichenposition
// vom CHAR_Array in das VRAM
procedure CHAR_Array_to_VRAM(X_PIXEL_POS, Y_PIXEL_POS : Word);
var
  Adresse : Word;
  x_cnt : Byte;
  y_cnt : Byte;
  Array_Pos : Word;
begin

  // Zeichen aus Puffer lesen -> in VRAM schreiben
  ClearBit(A16);
  ClearBit(DIR);

  DDRA:=  %11111111;

  // Sonderfall: Y_KUR_POS=12
  if Y_PIXEL_POS=121 then

    // Berechnung der Adresse (1.Pixel)
    Adresse:=Y_PIXEL_POS*512;
    Adresse:=Adresse+X_PIXEL_POS;
    ClearBit(A16);
    Array_Pos:=0;

    for y_cnt:=0 to 6 do

      for x_cnt:=0 to Zeichenbreite0 do

        PORTA:= CHAR_Array[Array_Pos];
        PORTB:= Low(Adresse);
        PORTC:= High(Adresse);

        ClearBit(WE);
        INC(Adresse);
        INC(Array_Pos);
        SetBit(WE);

      endfor;

      Adresse:=Adresse+512;
      Adresse:=Adresse-Zeichenbreite;

    endfor;

    // ... dann Y_PIX_POS 128

    // Berechnung der Adresse (1.Pixel)
    Adresse:=X_PIXEL_POS;
    SetBit(A16);

    for x_cnt:=0 to Zeichenbreite0 do

      PORTA:= CHAR_Array[Array_Pos];
      PORTB:= Low(Adresse);
      PORTC:= High(Adresse);

      ClearBit(WE);
      INC(Adresse);
      INC(Array_Pos);
      SetBit(WE);

    endfor; { for x_cnt:=0 to 5 do }

  else

    if Y_PIXEL_POS>121 then
      // Berechnung der Adresse (1.Pixel)
      Adresse:=Y_PIXEL_POS-128;
      Adresse:=Adresse*512;
      Adresse:=Adresse+X_PIXEL_POS;
      SetBit(A16);
    else
      // Berechnung der Adresse (1.Pixel)
      Adresse:=Y_PIXEL_POS*512;
      Adresse:=Adresse+X_PIXEL_POS;
      ClearBit(A16);
    endif; { if Y_KUR_POS>12 then }

    Array_Pos:=0;

    for y_cnt:=0 to 7 do

      for x_cnt:=0 to Zeichenbreite0 do

        PORTA:= CHAR_Array[Array_Pos];
        PORTB:= Low(Adresse);
        PORTC:= High(Adresse);

        ClearBit(WE);
        INC(Adresse);
        INC(Array_Pos);
        SetBit(WE);

      endfor; { for x_cnt:=0 to 5 do }

      Adresse:=Adresse+512;
      Adresse:=Adresse-Zeichenbreite;

    endfor; { for y_cnt:=0 to 7 do }

  endif; { else Y_KUR_POS=12 then }

  // Grundeinstellungen
  SetBit(WE);
  DDRA:= %00000000;
  SetBit(DIR);
  SetBit(OE);
  ClearBit(A16);


end CHAR_Array_to_VRAM;






















// *** Hilfsprogramme fuer CONOUT ***




procedure CURSOR_UP;
begin
  if Y_KUR_POS>0 then
    DEC(Y_KUR_POS);
    Y_PIX_POS:=Y_PIX_POS-Zeichenhoehe;
  endif;
end CURSOR_UP;







// Loeschen von Kursorpossition bis zum Zeilenende
procedure DELETE_EOL;
var
  Kur_Pos_Cnt : Word;
  X_KUR_POS2 : Word;
  X_PIX_POS2 : Word;
begin

  Zeichen:=$20;

  X_KUR_POS2:=X_KUR_POS;
  X_PIX_POS2:=X_PIX_POS;

  for Kur_Pos_Cnt:=X_KUR_POS2 to max_Spalten do
    Write_Char;
    INC(X_KUR_POS);
    X_PIX_POS:= X_PIX_POS+Zeichenbreite;
  endfor; { for Kur_Pos_Cnt:=X_KUR_POS2 to 79 do }

  if KopieBWS then

    Umschalten_Speicherseite;

    X_KUR_POS:=X_KUR_POS2;
    X_PIX_POS:=X_PIX_POS2;

    for Kur_Pos_Cnt:=X_KUR_POS2 to max_Spalten do
      Write_Char;
      INC(X_KUR_POS);
      X_PIX_POS:= X_PIX_POS+Zeichenbreite;
    endfor; { for Kur_Pos_Cnt:=X_KUR_POS2 to 79 do }

    Umschalten_Speicherseite;

  endif; { if KopieBeideSeiten then }

  X_KUR_POS:=X_KUR_POS2;
  X_PIX_POS:=X_PIX_POS2;

end DELETE_EOL;






// Loeschen von Anfang bis Kursorpossition
procedure DELETE_BOL;
var
  Kur_Pos_Cnt : Word;
  X_KUR_POS2 : Word;
  X_PIX_POS2 : Word;

begin

  Zeichen:=$20;

  X_KUR_POS2:=X_KUR_POS;
  X_PIX_POS2:=X_PIX_POS;

  X_KUR_POS:=0;
  X_PIX_POS:=X_ANF;

  for Kur_Pos_Cnt:=0 to X_KUR_POS2 do
    Write_Char;
    INC(X_KUR_POS);
    X_PIX_POS:= X_PIX_POS+Zeichenbreite;
  endfor; { for Kur_Pos_Cnt:=0 to X_KUR_POS2 do }

  if KopieBWS then

    Umschalten_Speicherseite;

    X_KUR_POS:=0;
    X_PIX_POS:=X_ANF;

    for Kur_Pos_Cnt:=0 to X_KUR_POS2 do
      Write_Char;
      INC(X_KUR_POS);
      X_PIX_POS:= X_PIX_POS+Zeichenbreite;
    endfor; { for Kur_Pos_Cnt:=0 to X_KUR_POS2 do }

    Umschalten_Speicherseite;

  endif; { if KopieBeideSeiten then }

  X_KUR_POS:=X_KUR_POS2;
  X_PIX_POS:=X_PIX_POS2;

end DELETE_BOL;







// Loeschen der gesammten Zeile auf Y-Kursorposition
procedure DELETE_LINE;
var
  Kur_Pos_Cnt : Word;
  X_KUR_POS2 : Word;
  X_PIX_POS2 : Word;

begin

  Zeichen:=$20;

  X_KUR_POS2:=X_KUR_POS;
  X_PIX_POS2:=X_PIX_POS;

  X_KUR_POS:=0;
  X_PIX_POS:=X_ANF;

  for Kur_Pos_Cnt:=0 to max_Spalten do
    Write_Char;
    INC(X_KUR_POS);
    X_PIX_POS:= X_PIX_POS+Zeichenbreite;
  endfor; { for Kur_Pos_Cnt:=0 to max_Spalten do }

  if KopieBWS then

    Umschalten_Speicherseite;

    X_KUR_POS:=0;
    X_PIX_POS:=X_ANF;

    for Kur_Pos_Cnt:=0 to max_Spalten do
      Write_Char;
      INC(X_KUR_POS);
      X_PIX_POS:= X_PIX_POS+Zeichenbreite;
    endfor; { for Kur_Pos_Cnt:=0 to max_Spalten do }

    Umschalten_Speicherseite;

  endif; { if KopieBeideSeiten then }


  X_KUR_POS:=X_KUR_POS2;
  X_PIX_POS:=X_PIX_POS2;

end DELETE_LINE;









// Loeschen der gesammten Zeile auf Y-Kursorposition
procedure CLEAR_LINE_on_RAMPAGE;
var
  Kur_Pos_Cnt : Word;
  X_KUR_POS2 : Word;
  X_PIX_POS2 : Word;

begin

  Zeichen:=$20;

  X_KUR_POS2:=X_KUR_POS;
  X_PIX_POS2:=X_PIX_POS;

  X_KUR_POS:=0;
  X_PIX_POS:=X_ANF;

  for Kur_Pos_Cnt:=0 to max_Spalten2 do
    Write_Char;
    INC(X_KUR_POS);
    X_PIX_POS:= X_PIX_POS+Zeichenbreite;
  endfor; { for Kur_Pos_Cnt:=0 to 79 do }

  X_KUR_POS:=X_KUR_POS2;
  X_PIX_POS:=X_PIX_POS2;

end CLEAR_LINE_on_RAMPAGE;












// Loeschen der letzten Zeile
procedure DELETE_last_LINE_on_RAMPAGE;
var
  Kur_Pos_Cnt : Word;
  X_KUR_POS2 : Word;
  X_PIX_POS2 : Word;
  Y_KUR_POS2 : Word;
  Y_PIX_POS2 : Word;

begin

  Zeichen:=$20;

  X_KUR_POS2:=X_KUR_POS;
  X_PIX_POS2:=X_PIX_POS;
  Y_KUR_POS2:=Y_KUR_POS;
  Y_PIX_POS2:=Y_PIX_POS;

  X_KUR_POS:=0;
  X_PIX_POS:=X_ANF;
  Y_KUR_POS:=max_Zeilen;
  Y_PIX_POS:=240-Zeichenhoehe;
  Y_PIX_POS:=Y_PIX_POS+Y_ANF; // 231

  for Kur_Pos_Cnt:=0 to max_Spalten do
    Write_Char;
    INC(X_KUR_POS);
    X_PIX_POS:= X_PIX_POS+Zeichenbreite;
  endfor; { for Kur_Pos_Cnt:=0 to 79 do }

  X_KUR_POS:=X_KUR_POS2;
  X_PIX_POS:=X_PIX_POS2;
  Y_KUR_POS:=Y_KUR_POS2;
  Y_PIX_POS:=Y_PIX_POS2;

end DELETE_last_LINE_on_RAMPAGE;








// Loeschen der aktuellen Zeile auf der Kursorposition
// und ScreenRoll
procedure DELETE_LINE_SCR_ROLL_on_RAMPAGE;
var
  Kur_Pos_X_Cnt : Word;
  Kur_Pos_Y_Cnt : Word;
  X_PIXEL_POS : Word;
  Y_PIX_POS_von : Word;
  Y_PIX_POS_nach : Word;

begin

  if Y_KUR_POS<=max_Zeilen2 then
    for Kur_Pos_Y_Cnt:=Y_KUR_POS to max_Zeilen2 do

      Y_PIX_POS_nach:=Kur_Pos_Y_Cnt*Zeichenhoehe;
      Y_PIX_POS_nach:=Y_PIX_POS_nach+Y_ANF;
      Y_PIX_POS_von:=Y_PIX_POS_nach+Zeichenhoehe;

      X_PIXEL_POS:=X_ANF;

      for Kur_Pos_X_Cnt:=0 to max_Spalten do
        VRAM_to_CHAR_Array(X_PIXEL_POS,Y_PIX_POS_von);
        CHAR_Array_to_VRAM(X_PIXEL_POS,Y_PIX_POS_nach);
        X_PIXEL_POS:= X_PIXEL_POS+Zeichenbreite;
      endfor; { for Kur_Pos_X_Cnt:=0 to max_Spalten do }

    endfor; { for Kur_Pos_Y_Cnt:=Y_KUR_POS to max_Zeilen2 do }

  endif; { if Y_KUR_POS<=22 then }

end DELETE_LINE_SCR_ROLL_on_RAMPAGE;





// Loeschen der aktuellen Zeile auf der Kursorposition
// und ScreenRoll
procedure DELETE_LINE_SCR_ROLL;
begin
  DELETE_LINE_SCR_ROLL_on_RAMPAGE;
  DELETE_last_LINE_on_RAMPAGE;

  if KopieBWS then

    Umschalten_Speicherseite;

    DELETE_LINE_SCR_ROLL_on_RAMPAGE;
    DELETE_last_LINE_on_RAMPAGE;
    Umschalten_Speicherseite;

  endif; { if KopieBeideSeiten then }

end DELETE_LINE_SCR_ROLL;












// neue Zeile der aktuellen Zeile auf der Kursorposition
// und ScreenRoll
procedure INSERT_LINE_SCR_ROLL_on_RAMPAGE;
var
  Kur_Pos_X_Cnt : Word;
  Kur_Pos_Y_Cnt : Word;
  X_PIXEL_POS : Word;
  Y_PIX_POS_von : Word;
  Y_PIX_POS_nach : Word;

begin

  if Y_KUR_POS<=max_Zeilen2 then
    for Kur_Pos_Y_Cnt:=max_Zeilen2 downto Y_KUR_POS do

      Y_PIX_POS_von:=Kur_Pos_Y_Cnt*Zeichenhoehe;
      Y_PIX_POS_von:=Y_PIX_POS_von+Y_ANF;
      Y_PIX_POS_nach:=Y_PIX_POS_von+Zeichenhoehe;

      X_PIXEL_POS:=X_ANF;

      for Kur_Pos_X_Cnt:=0 to max_Spalten do
        VRAM_to_CHAR_Array(X_PIXEL_POS,Y_PIX_POS_von);
        CHAR_Array_to_VRAM(X_PIXEL_POS,Y_PIX_POS_nach);
        X_PIXEL_POS:= X_PIXEL_POS+Zeichenbreite;
      endfor; { for Kur_Pos_X_Cnt:=0 to max_Spalten do }

    endfor; { for Kur_Pos_Y_Cnt:=max_Zeilen2 downto Y_KUR_POS do }

  endif; { if Y_KUR_POS<=max_Zeilen2 then }

end INSERT_LINE_SCR_ROLL_on_RAMPAGE;









// neue Zeile der aktuellen Zeile auf der Kursorposition
// und ScreenRoll
procedure INSERT_LINE_SCR_ROLL;
begin
  INSERT_LINE_SCR_ROLL_on_RAMPAGE;
  CLEAR_LINE_on_RAMPAGE;

  if KopieBWS then

    Umschalten_Speicherseite;

    INSERT_LINE_SCR_ROLL_on_RAMPAGE;
    CLEAR_LINE_on_RAMPAGE;
    Umschalten_Speicherseite;

  endif; { if KopieBeideSeiten then }

end INSERT_LINE_SCR_ROLL;







procedure ERASE_DOWN_on_RAMPAGE;
var
  Kur_Pos_X_Cnt : Byte;
  Kur_Pos_Y_Cnt : Word;
  X_KUR_POS2 : Word;
  X_PIX_POS2 : Word;
  Y_KUR_POS2 : Word;
  Y_PIX_POS2 : Word;

begin

  Zeichen:=$20;

  X_KUR_POS2:=X_KUR_POS;
  X_PIX_POS2:=X_PIX_POS;

  Y_KUR_POS2:=Y_KUR_POS;
  Y_PIX_POS2:=Y_PIX_POS;

  for Y_KUR_POS:=Y_KUR_POS2 to max_Zeilen do

    X_PIX_POS:=X_ANF;

    for X_KUR_POS:=0 to max_Spalten2 do
      Write_Char;
      X_PIX_POS:=X_PIX_POS+Zeichenbreite;
    endfor; { for Kur_Pos_Cnt:=0 to 79 do }

    Y_PIX_POS:=Y_PIX_POS+Zeichenhoehe;

  endfor;

  X_KUR_POS:=X_KUR_POS2;
  X_PIX_POS:=X_PIX_POS2;

  Y_KUR_POS:=Y_KUR_POS2;
  Y_PIX_POS:=Y_PIX_POS2;

end ERASE_DOWN_on_RAMPAGE;



// Erases the screen from the current line down to the bottom of the screen.
procedure ERASE_DOWN;
begin
  ERASE_DOWN_on_RAMPAGE;

  if KopieBWS then

    Umschalten_Speicherseite;

    ERASE_DOWN_on_RAMPAGE;
    Umschalten_Speicherseite;

  endif; { if KopieBeideSeiten then }

end ERASE_DOWN;











procedure SET_CURSOR_POSITION(Zeile, Spalte : Word); // bezogen auf 0,0 !!!
begin

  // 1. Zeile setzen
  X_KUR_POS:=Spalte;

  if X_KUR_POS>max_Spalten then
    X_KUR_POS:=max_Spalten;
  endif;

  X_PIX_POS:=X_KUR_POS*Zeichenbreite;
  X_PIX_POS:=X_PIX_POS+X_ANF;


  // 2. Spalte setzen
  Y_KUR_POS:=Zeile;

  if Y_KUR_POS>max_Zeilen then
    Y_KUR_POS:=max_Zeilen;
  endif;

  Y_PIX_POS:=Y_KUR_POS*Zeichenhoehe;
  Y_PIX_POS:=Y_PIX_POS+Y_ANF;

end SET_CURSOR_POSITION; 



// Erhoeht die Kursorposition der aktuellen Zeile um 1 
//     (wenn die max_Spalten noch nicht erreicht ist) 
procedure CURSOR_RIGHT;
begin

  if X_KUR_POS<max_Spalten then
    Inc(X_KUR_POS);
    X_PIX_POS:=X_PIX_POS+Zeichenbreite;
  endif;

end CURSOR_RIGHT;



procedure Write_Char_in_beide_RP;
begin
  Write_Char;

  if KopieBWS then

    Umschalten_Speicherseite;
    Write_Char;
    Umschalten_Speicherseite;

  endif; { if KopieBeideSeiten then }

  CURSOR_RIGHT;
end Write_Char_in_beide_RP;




// *********************

// *** TERMINAL MODE ***

// *********************









procedure Normal_Terminal_Mode;
begin


  case Zeichen of

    $0A :   // 0A = Zeilenschaltung (Line Feed)
            // Y_KUR_POS von 0..23 = 24 Zeilen pro Bildschirm
            if Y_KUR_POS<max_Zeilen then
              INC(Y_KUR_POS);
              Y_PIX_POS:= Y_PIX_POS+Zeichenhoehe;
            else
              ScreenRoll;
              if KopieBWS then
                Umschalten_Speicherseite;
                ScreenRoll;
                Umschalten_Speicherseite;
              endif; { if KopieBeideSeiten then }


            endif;
 
    |

   $0C :    // 0C = Bildschirm loeschen (Clear Screen)
            clear_RAM;
            if KopieBWS then
              Umschalten_Speicherseite;
              clear_RAM;
              Umschalten_Speicherseite;
            endif; { if KopieBeideSeiten then }


 
    |

    $0D :   // 0D = Wagenruecklauf (Carridge Return)
            X_KUR_POS:=0;
            X_PIX_POS:=X_ANF;
 
    | 

    $1B :   // ESC-Modus
            TermMode:=IBM_ESC_TM;

    | 

  else


    if Zeichen>=$20 then
      if Zeichen<=$7F then
        if X_KUR_POS<=max_Spalten then

          Write_Char_in_beide_RP;

          VRAM_to_Kursor_Array;
          Kursor_sichtbar:=false;
          Kursor_Schleife:=0;
          KurArr_Aktuell:=true;


        endif; { if X_KUR_POS<80 then } 
      endif; { if Zeichen<=$7F then } 
    endif; { if Zeichen>=$20 then  }


  endcase; { case Zeichen of }



end Normal_Terminal_Mode;
  
  





procedure Escape_Reset;
begin
  CSI_Parameter1:=0;
  CSI_Parameter2:=0;
  CSI_Parameter3:=0;
  TermMode:=Normal_TM;
  Escape_Sequenz:='';
end Escape_Reset;






procedure Escape_Fehler;
var
  ESC_Zaehler : Byte;
  ESC_Laenge  : Byte;
  ESC_Zeichen : Char;
begin

  Zeichen:=$45; // E
  Write_Char_in_beide_RP;

  Zeichen:=$53; // S
  Write_Char_in_beide_RP;

  Zeichen:=$43; // C
  Write_Char_in_beide_RP;

  ESC_Laenge:=Length(Escape_Sequenz);

  for ESC_Zaehler:=1 to ESC_Laenge do
    ESC_Zeichen:=Escape_Sequenz[ESC_Zaehler];
    Zeichen:=CharToByte(ESC_Zeichen);
    Write_Char_in_beide_RP;
  endfor;

  Escape_Reset;
end Escape_Fehler;




procedure IBM_ESC_Terminal_Mode;
begin

  case Zeichen of

    $5B : TermMode:=IBM_CSI_TM; // Control Sequence Introducer (CSI) = ESC [ 
    |

    $47 : TermMode:=IBM_GRAFIK_Par_1_TM; // GRAFIK-Mode = ESC G
    |

  else

    Escape_Fehler;

  endcase;

end IBM_ESC_Terminal_Mode;














procedure Set_Display_Attributes(DispAttr : Word);
begin


  case DispAttr of

    // *** Textattribute ***


    // ESC [ 0 m = Alle Attribute AUS (Reset all attributes)
    0 : FCOLOR:=cl_Blue;
    |

    // ESC [ 1 m = Fett ein
    1 : FCOLOR:=cl_Gray;
    |

    // ESC [ 4 m = Unterstreichen
    4 : FCOLOR:=cl_Purple;
    |

    // ESC [ 5 m = Blinken EIN
    5 : FCOLOR:=cl_Red;
    |

    // ESC [ 7 m = Invertierte Darstellung
    7 : FCOLOR:=cl_Black;
    |

    // ESC [ 8 m = Versteckt ein
    8 : FCOLOR:=cl_White;
    |



   // *** Schriftgroesse ***

    // ESC [ 10 m = Font 6x8
    10 :   setze_Font(fnt_6x8);
    |

    // ESC [ 11 m = Font 8x8
    11 :   setze_Font(fnt_8x8);
    |



    // *** Vordergrundfarben ***

    // ESC [ 30 m = Vordergrundfarbe - 30 Schwarz
    30 : FCOLOR:=cl_Black;
    |

    // ESC [ 31 m = Vordergrundfarbe - 31 Rot
    31 : FCOLOR:=cl_Red;
    |

    // ESC [ 32 m = Vordergrundfarbe - 32 Grn
    32 : FCOLOR:=cl_Green;
    |

    // ESC [ 33 m = Vordergrundfarbe - 33 Gelb
    33 : FCOLOR:=cl_Yellow;
    |

    // ESC [ 34 m = Vordergrundfarbe - 34 Blau
    34 : FCOLOR:=cl_Blue;
    |

    // ESC [ 35 m = Vordergrundfarbe - 35 Karmesinrot (Purpur, lila)
    35 : FCOLOR:=cl_Purple;
    |

    // ESC [ 36 m = Vordergrundfarbe - 36 Cyanblau
    36 : FCOLOR:=cl_Aqua;
    |

    // ESC [ 37 m = Vordergrundfarbe - 37 Wei
    37 : FCOLOR:=cl_White;
    |



    // *** Hintergrundfarben ***

    // ESC [ 40 m = Hintergrundfarbe - 40 Schwarz
    40 : BCOLOR:=cl_Black;
    |

    // ESC [ 41 m = Hintergrundfarbe - 41 Rot
    41 : BCOLOR:=cl_Red;
    |

    // ESC [ 42 m = Hintergrundfarbe - 42 Grn
    42 : BCOLOR:=cl_Green;
    |

    // ESC [ 43 m = Hintergrundfarbe - 43 Gelb
    43 : BCOLOR:=cl_Yellow;
    |

    // ESC [ 44 m = Hintergrundfarbe - 44 Blau
    44 : BCOLOR:=cl_Blue;
    |

    // ESC [ 45 m = Hintergrundfarbe - 45 Karmesinrot (Purpur, lila)
    45 : BCOLOR:=cl_Purple;
    |

    // ESC [ 46 m = Hintergrundfarbe - 46 Cyanblau
    46 : BCOLOR:=cl_Aqua;
    |

    // ESC [ 47 m = Hintergrundfarbe - 47 Wei
    47 : BCOLOR:=cl_White;
    |

    else
      Escape_Fehler;
    endcase;

end Set_Display_Attributes;




















procedure IBM_CSI_Terminal_Mode;
begin

  case Zeichen of

    $30..$39 : CSI_Parameter1:=ByteToWord(Zeichen)-$30;
               TermMode:=IBM_CSI_TM_Par_1;
    |

    $4B : // Loeschen bis Zeilenende = ESC [ K
          DELETE_EOL; 
          Escape_Reset;
    |

    $66 : // Schirm loeschen = ESC [ f
          Reset_Form;
          Escape_Reset;
    |

    $6D : // Turn off character attributes = ESC [ m
          Set_Display_Attributes(0);
          Escape_Reset;
    |

    $3F : // Hide/Show the cursor = ESC [ ? 2 5 h
          CSI_Parameter1:=0;
          TermMode:=IBM_CSI_TM_Par_1;
    |

    $48 : // "H" -- MOVE CURSOR = ESC [ H
          SET_CURSOR_POSITION(0,0);
          Escape_Reset;
    |

    $4A : // "J" -- Erase Down = ESC [ J
          // Erases the screen from the current line down to the bottom of the screen.
          ERASE_DOWN;
          Escape_Reset;
    |

    $4D : // "M" -- Move the cursor up in scrolling region = ESC [ M
          DELETE_LINE;
          CURSOR_UP;
          DELETE_LINE;
          Escape_Reset;
    |

  else

    Escape_Fehler;
 
  endcase; { case Zeichen of }

end IBM_CSI_Terminal_Mode;










// IBM_CSI_TM_Par_1
procedure IBM_CSI_Par_1_Terminal_Mode;
var
  Zaehler : Word;
begin

  case Zeichen of

    $30..$39 : CSI_Parameter1:=CSI_Parameter1*10;
               CSI_Parameter1:=CSI_Parameter1+ByteToWord(Zeichen);
               CSI_Parameter1:=CSI_Parameter1-$30;

               TermMode:=IBM_CSI_TM_Par_1;
    |

    $3B :      // ";"
	       // ESC [ Pn1 ; PN2 H oder
	       // ESC [ Pn1 ; PN2 f oder
               // ESC [ Pn1 ; PN2 m oder
	       // Xterm 256 colors - ESC[48;5;BCOLOR m oder
	       // Xterm 256 colors - ESC[38;5;FCOLOR m
               TermMode:=IBM_CSI_TM_Par_2;

    |

    $4A : // Zeilen loeschen = ESC [ Pn J

          case CSI_Parameter1 of
 
            2 : // Loeschen gesammte zeile
                Reset_Form;

            |
          else
            Escape_Fehler;
          endcase;
          Escape_Reset;
    |


    $4B : // Zeilen loeschen = ESC [ Pn K

          case CSI_Parameter1 of

            0 : // Lschen bis Zeilenende
                DELETE_EOL; 
            |

            1 : // Lschen von Anfang bis Kursor
                DELETE_BOL; 
            |

            2 : // Lschen gesammte zeile
                DELETE_LINE;
            |

          else
            Escape_Fehler;
          endcase;
          Escape_Reset;
    |


    $4C : // INSERT_LINE_SCR_ROLL = ESC [ Pn L
          for Zaehler:=0 to CSI_Parameter1-1 do
            INSERT_LINE_SCR_ROLL;
          endfor;
          Escape_Reset;
    |


    $4D : // "M" DELETE_LINE_SCR_ROLL = ESC [ Pn M
          for Zaehler:=0 to CSI_Parameter1-1 do
            DELETE_LINE_SCR_ROLL;
          endfor;
          Escape_Reset;
    |


    $6D : // "m" Character Attributes = ESC [ Pn m
          Set_Display_Attributes(CSI_Parameter1);
          Escape_Reset;
    |



    $6C : // "l" ESC [ ? X X l

          case CSI_Parameter1 of

            // "l" Hide the cursor = ESC [ ? 2 5 l (Hide the cursor: 0x9B 0x3F 0x32 0x35 0x6C)
            25 : if Kursor_sichtbar then
                   Kursor_sichtbar:=false;
                   Schalte_Kursor_Aus;
                 endif;

                 Kursor_Ein:=false;
                 Escape_Reset;

            |

            // "l" KopieBeideSeiten AUS = ESC [ ? 2 6 l
            26 : KopieBWS:=false;
                 Escape_Reset;

            |

          else
            Escape_Fehler;
          endcase; { case CSI_Parameter1 of }

    |


    $68 : // "h" ESC [ ? X X h

          case CSI_Parameter1 of

            // "h" Show the cursor = ESC [ ? 2 5 h (Show the cursor: 0x9B 0x3F 0x32 0x35 0x68)
            25 : Kursor_Schleife:=0;
                 Kursor_Ein:=true;
                 Escape_Reset;

            |

            // "h" KopieBeideSeiten EIN = ESC [ ? 2 6 h
            26 : KopieBWS:=true;
                 Escape_Reset;

            |


            // "h" UmschaltenSpeicherseite = ESC [ ? 2 7 h
            27 : Umschalten_Speicherseite;
                 Escape_Reset;

            |

          else
            Escape_Fehler;
          endcase; { case CSI_Parameter1 of }

    |

  else

    Escape_Fehler;

  endcase; { case Zeichen of }

end IBM_CSI_Par_1_Terminal_Mode;








// IBM_CSI_TM_Par_2
procedure IBM_CSI_Par_2_Terminal_Mode;
begin

  case Zeichen of

    $30..$39 : CSI_Parameter2:=CSI_Parameter2*10;
               CSI_Parameter2:=CSI_Parameter2+ByteToWord(Zeichen);
               CSI_Parameter2:=CSI_Parameter2-$30;

               TermMode:=IBM_CSI_TM_Par_2;
    |

    $3B : // ";" 
          // Xterm 256 colors - ESC[48;5;BCOLOR m
          // Xterm 256 colors - ESC[38;5;FCOLOR m
          TermMode:=IBM_CSI_TM_Par_3;
    |


    $48 : // "H" -- MOVE CURSOR = ESC [ Pn1 ; PN2 H
          SET_CURSOR_POSITION(CSI_Parameter1,CSI_Parameter2);
          Escape_Reset;
    |

    $66 : // "f" -- MOVE CURSOR = ESC [ Pn1 ; PN2 f
          SET_CURSOR_POSITION(CSI_Parameter1,CSI_Parameter2);
          Escape_Reset;
    |

    $6D : // "m" -- 2 mal Character Attributes = ESC [ Pn1 ; PN2 m
          Set_Display_Attributes(CSI_Parameter1);
          Set_Display_Attributes(CSI_Parameter2);
          Escape_Reset;

    |
 
  else
 
    Escape_Fehler;
 
  endcase; { case Zeichen of }


end IBM_CSI_Par_2_Terminal_Mode;





procedure IBM_CSI_Par_3_Terminal_Mode;
begin


  case Zeichen of

    $30..$39 : CSI_Parameter3:=CSI_Parameter3*10;
               CSI_Parameter3:=CSI_Parameter3+ByteToWord(Zeichen);
               CSI_Parameter3:=CSI_Parameter3-$30;

               TermMode:=IBM_CSI_TM_Par_3;

    |


    $6D : // "m"

          case CSI_Parameter1 of

          48 : // Xterm 256 colors - ESC[48;5;BCOLOR m
               if CSI_Parameter2=5 then
                 BCOLOR:=Low(CSI_Parameter3);
               else
                 Escape_Fehler;
               endif;
          |

          38 : // Xterm 256 colors - ESC[38;5;FCOLOR m
               if CSI_Parameter2=5 then
                 FCOLOR:=Low(CSI_Parameter3);
               else
                 Escape_Fehler;
               endif;
          |

          else
            Set_Display_Attributes(CSI_Parameter1);
            Set_Display_Attributes(CSI_Parameter2);
            Set_Display_Attributes(CSI_Parameter3);
          endcase;

          Escape_Reset;

    |

  else

    Escape_Fehler;

  endcase; { case Zeichen of }


end IBM_CSI_Par_3_Terminal_Mode;





// ESC G - Auswertung der ESC-Grafik-Steuerzeichen



procedure IBM_GRAFIK_Par_1_Terminal_Mode;
begin

  case Zeichen of

    $30..$39 : CSI_Parameter1:=CSI_Parameter1*10;
               CSI_Parameter1:=CSI_Parameter1+ByteToWord(Zeichen);
               CSI_Parameter1:=CSI_Parameter1-$30;

               TermMode:=IBM_GRAFIK_Par_1_TM;

    |

    $3B : // ";"
          TermMode:=IBM_GRAFIK_Par_2_TM;

    |

    $43 : // "C" -- GrafikSetzeFarbe = ESC G Pn1 C
          PIXCOLOR:=Low(CSI_Parameter1);
          Escape_Reset;

    |

    $4B : // "K" -- GrafikZeichneKreis = ESC G Pn1 K
          GrafikZeichneKreis(CSI_Parameter1);
          Escape_Reset;

    |

    $56 : // "V" -- GrafikZeichneVollenKreis = ESC G Pn1 V
          GrafikZeichneVollenKreis(CSI_Parameter1);
          Escape_Reset;

    |


  else

    Escape_Fehler;

  endcase; { case Zeichen of }


end IBM_GRAFIK_Par_1_Terminal_Mode;





procedure IBM_GRAFIK_Par_2_Terminal_Mode;
begin

  case Zeichen of

    $30..$39 : CSI_Parameter2:=CSI_Parameter2*10;
               CSI_Parameter2:=CSI_Parameter2+ByteToWord(Zeichen);
               CSI_Parameter2:=CSI_Parameter2-$30;

               TermMode:=IBM_GRAFIK_Par_2_TM;

    |

    $47 : // "G" -- GrafikSetCursor = ESC G Pn1 ; Pn2 G
          X_PIX_POS:=CSI_Parameter1;
          Y_PIX_POS:=CSI_Parameter2;
          Escape_Reset;

    |

    $50 : // "P" -- GrafikDrawPoint = ESC G Pn1 ; Pn2 P
          X_PIX_POS:=CSI_Parameter1;
          Y_PIX_POS:=CSI_Parameter2;
          GrafikDrawPoint;
          Escape_Reset;

    |

    $4C : // "L" -- GrafikDrawLine = ESC G Pn1 ; Pn2 L
          X_END_POS:=CSI_Parameter1;
          Y_END_POS:=CSI_Parameter2;
          GrafikDrawLine;
          Escape_Reset;

    |

    $52 : // "R" -- GrafikDrawRectangle = ESC G Pn1 ; Pn2 R
          X_END_POS:=CSI_Parameter1;
          Y_END_POS:=CSI_Parameter2;
          GrafikDrawRectangle;
          Escape_Reset;

    |

    $46 : // "F" -- GrafikDrawFullRectangle = ESC G Pn1 ; Pn2 F
          X_END_POS:=CSI_Parameter1;
          Y_END_POS:=CSI_Parameter2;
          GrafikDrawFullRectangle;
          Escape_Reset;

    |

  else

    Escape_Fehler;

  endcase; { case Zeichen of }


end IBM_GRAFIK_Par_2_Terminal_Mode;








procedure CONOUT;
var
  C_Zeichen : Char;
begin

  if Kursor_sichtbar then
    Schalte_Kursor_Aus;
    Kursor_sichtbar:=false;
  endif;
  Kursor_Schleife:=0;
  KurArr_Aktuell:=false;

  // Abspeichern der Escape-Steuerzeichen fuer die
  // Anzeige der nicht ausgewerteter Sequenzen im Fehlerfall
  if TermMode<>Normal_TM then
    C_Zeichen:=ByteToChar(Zeichen);
    Escape_Sequenz:=Escape_Sequenz+C_Zeichen;
  endif;
 
  case TermMode of

    Normal_TM            : Normal_Terminal_Mode;
    |

    IBM_ESC_TM           : IBM_ESC_Terminal_Mode;
    |

    IBM_CSI_TM           : IBM_CSI_Terminal_Mode;
    |

    IBM_CSI_TM_Par_1     : IBM_CSI_Par_1_Terminal_Mode;
    |

    IBM_CSI_TM_Par_2     : IBM_CSI_Par_2_Terminal_Mode;
    |

    IBM_CSI_TM_Par_3     : IBM_CSI_Par_3_Terminal_Mode;
    |

    IBM_GRAFIK_Par_1_TM  : IBM_GRAFIK_Par_1_Terminal_Mode;
    |

    IBM_GRAFIK_Par_2_TM  : IBM_GRAFIK_Par_2_Terminal_Mode;
    |

  else

    Escape_Fehler;

  endcase;

end CONOUT;











// ******************************
// *** serielle Schnittstelle ***
// ******************************


procedure sende_Zeichen(s_Zeichen : Byte);
begin
  // wenn UDRE0 gesetzt ist, so ist der Puffer leer
  // -> warten solange UDRE0 = 0  
  while BitLow(UDRE0) do
    SetBit(LED_GN); // gruen ein
  endwhile;

  ClearBit(LED_GN); // gruen aus

  UDR0:=s_Zeichen;

end sende_Zeichen; 



procedure RX_Puffer_abfragen;
begin
  // wurde ein Byte durch die UART empfangen, so ist das RXC0-Bit gesetzt 
  // ist RXC0-Bit = 0, dann ist der Empfangs-Puffer leer
  if BitHigh(RXC0) then
    Zeichen:=UDR0;
    CONOUT;
    sende_Zeichen(XON);
  endif; { if BitHigh(RXC0) then }
end RX_Puffer_abfragen;







{--------------------------------------------------------------}
{ Main Program }
{$IDATA}







begin

ASM;

.equ UBRR0 = UBRR0L

ENDASM;

  setze_Font(fnt_6x8);
  KopieBWS:=true;

  Kursor_Ein:=true;
  Kursor_sichtbar:=false;
  KurArr_Aktuell:=false;
  Kursor_Schleife:=0;

  Escape_Sequenz:='';

  TermMode:=Normal_TM;
  CSI_Parameter1:=0;
  CSI_Parameter2:=0;
  CSI_Parameter3:=0;


  InitPorts;
  INIT_SER_PORT;
  Init_BILD_Adresse;

  ClearBit(LED_GN);
  ClearBit(VS1_GS2);

  // Grundeinstellung
  DDRA:=  %00000000;
  SetBit(WE);
  SetBit(DIR);
  SetBit(OE);

  PORTA:=%00000000;
  PORTB:=%00000000;
  PORTC:=%00000000;
  ClearBit(A16);

  Reset_Form;

  BILD_Adresse:=ArrToAdress(BILD5_256);
  kopiere_Hintergrundbild;

  kopiere_Kreis;

  sende_Zeichen(XON);


  loop

    RX_Puffer_abfragen;

    Inc(Kursor_Schleife);

    if Kursor_Ein then
      if Kursor_Schleife=63000 then
        Kursor_Schleife:=0;
        if Kursor_sichtbar then
          Kursor_sichtbar:=false;
          Schalte_Kursor_Aus;
        else
          Kursor_sichtbar:=true;
          Schalte_Kursor_Ein;
        endif; { if Kursor_Ein then }
      endif; { if Kursor_Schleife=63000 then }
    endif; { if Kursor_Ein then }

  endloop;


end GCON.


