program AE;


Device = mega162;

{$I mega162_Register.INC}


const

  // Befehlscode fuer den Adress Extender
  kein_Code       : Byte = 00;
  LOAD_ADRESS     : Byte = 01;   // laden einer neuen Adresse
  INC_ADRESS      : Byte = 02;   // Inkrement der geladenen Adresse
  ADRESS_TRISTATE : Byte = 03;   // Adressbus auf Treestate setzen
  WAIT_Z80ToBuf   : Byte = 04;   // Wait Z80-Bus to AE-Buffer
  WAIT_BufToBC    : Byte = 05;   // Wait AE-Buffer to Bus-Controller


  BB_Laenge : Word = 900;  // 3*300

var

  adr_00_07, adr_08_15, adr_16_18 : Byte;
  BefehlsCode : Byte;
  ByteBuffer : Array [1..900] of Byte;
  ByteBufferCounter : Word;


{
PB0 = frei

PB1 = /AECMD   E <-
PB2 = /AECS    E <-
PB3 = /AEINCRQ A ->
PB4 = LED_GELB A ->

PB5 = MOSI
PB6 = MISO
PB7 = SCK 
}
  AECMD_E [PINB,1]    : bit;
  AECS_E [PINB,2]     : bit;
  AERQ_A [PortB,3]    : bit;
  LED_GELB_A [PortB,4]: bit;



{--------------------------------------------------------------}
{ functions }

procedure InitPorts;
begin
  // Initialisieren der Ausgabeports

  // PortA => PA0-PA7 (8 Daten) (Eingang) 
  DDRA:= %00000000;

{
PB0 = NN    
PB1 = /AECMD   E <-
PB2 = /AECS    E <-
PB3 = /AEINCRQ A ->
PB4 = LED_GELB A ->
PB5 = MOSI
PB6 = MISO
PB7 = SCK
}
  //      EEEAAEEE
  DDRB:= %00011000;
  PortB:=%00001000;

  // PortC => PC0-PC7 = (8 Adr 00-07) (Ausgang) 
  DDRC:= %11111111;
  PortC:=%00000000;

  // PortD => PD0-PD7 = (8 Adr 08-15) (Ausgang) 
  DDRD:= %11111111;
  PortD:=%00000000;

  // PortE => PE0-PE2 = (3 Adr 16-18) (Ausgang)
  DDRE:= %11111111;
  PortE:=%00000000;

end InitPorts;



procedure PRO_Start;
begin

  // die Schleife laeuft so lange der Befehlscode=ADRESS_TRISTATE ist
  while Befehlscode=kein_Code do

    // der Bus Controller stellt eine Anforderung an den AE
    if BitLow(AECS_E) then

      // ist AECMD_E auf High (inaktiv) so soll ein weiterer
      // Inkrement der Adresse folgen
      if BitLow(AECMD_E) then

        // ist AECMD_E auf Low (aktiv) so soll
        // ein neuer Befehl uebertragen werden
        // Einlesen des BefehlsCodes
        BefehlsCode:=PINA;

        if BefehlsCode=kein_Code then

          // der BC kann benachrichtigt werden
          SetBit(LED_GELB_A);
          ClearBit(AERQ_A);

          // warten bis der BC die Uebernahme quittiert
          while BitLow(AECS_E) do
          endwhile;

          // Abschliessen des Einlesens
          SetBit(AERQ_A);
          ClearBit(LED_GELB_A);

        endif; { if BefehlsCode=kein_Code then }

      else

        SetBit(LED_GELB_A);

      endif; { if BitLow(AECMD_E) then }

    endif; { if BitLow(AECS_E) then }

  endwhile; { while Befehlscode=kein_Code do }

end PRO_Start;






// Laden der 3 AdressBytes
procedure PRO_LOAD_ADRESS;
begin

  // PortA => (Eingang)
  DDRA:= %00000000;

  SetBit(LED_GELB_A);

  // der BC kann benachrichtigt werden
  SetBit(LED_GELB_A);
  ClearBit(AERQ_A);

  // warten bis der BC die Uebernahme quittiert
  while BitLow(AECS_E) do
  endwhile;

  // Abschliessen des Einlesens
  SetBit(AERQ_A);


  // 1. Einlesen von adr_00_07
  // =========================

  // warten, bis der BC die Adresse sendet
  while BitHigh(AECS_E) do
  endwhile;

  // Einlesen der Adresse 
  adr_00_07:=PINA;

  // der BC kann benachrichtigt werden
  ClearBit(AERQ_A); 

  // warten bis der BC die Uebernahme quittiert
  while BitLow(AECS_E) do
  endwhile;

  // Abschliessen des Einlesens
  SetBit(AERQ_A);


  // 2. Einlesen von adr_08_15
  // =========================

  // warten, bis der BC die Adresse sendet
  while BitHigh(AECS_E) do
  endwhile;

  // Einlesen der Adresse 
  adr_08_15:=PINA;

  // der BC kann benachrichtigt werden
  ClearBit(AERQ_A); 

  // warten bis der BC die Uebernahme quittiert
  while BitLow(AECS_E) do
  endwhile;

  // Abschliessen des Einlesens
  SetBit(AERQ_A);


  // 3. Einlesen von adr_16_18
  // =========================

  // warten, bis der BC die Adresse sendet
  while BitHigh(AECS_E) do
  endwhile;

  // Einlesen der Adresse 
  adr_16_18:=PINA;

  Befehlscode:=INC_ADRESS;

end PRO_LOAD_ADRESS;







procedure PRO_INC_ADRESS;
begin


  // PortC => PC0-PC7 = (8 Adr 00-07) (Ausgang) 
  DDRC:= %11111111;
  PORTC := adr_00_07;

  // PortD => PD0-PD7 = (8 Adr 08-15) (Ausgang) 
  DDRD:= %11111111;
  PORTD := adr_08_15;

  // PortE => PE0-PE2 = (3 Adr 16-18) (Ausgang)
  DDRE:= %11111111;
  PORTE := adr_16_18;


  ASM;

  ; Inkrementieren der internen Adresse
  LDS        R16, adr_00_07  ; 1 Cycle
  LDS        R17, adr_08_15  ; 1 Cycle
  LDS        R18, adr_16_18  ; 1 Cycle
  CLR        R19             ; 1 Cycle
  SEC                        ; 1 Cycle
  ADC        R16, R19        ; 1 Cycle
  ADC        R17, R19        ; 1 Cycle
  ADC        R18, R19        ; 1 Cycle
  STS        adr_00_07, R16  ; 1 Cycle
  STS        adr_08_15, R17  ; 1 Cycle
  STS        adr_16_18, R18  ; 1 Cycle
  ; Summe 11 Takte

  ENDASM; 


  // der BC kann benachrichtigt werden
  SetBit(LED_GELB_A);
  ClearBit(AERQ_A); 

  // warten bis der BC die Uebernahme quittiert
  while BitLow(AECS_E) do
  endwhile;

  // Abschliessen des Einlesens
  SetBit(AERQ_A);
  ClearBit(LED_GELB_A);


  // die Schleife laeuft so lange der Befehlscode=INC_ADRESS ist
  while Befehlscode=INC_ADRESS do

    // der Bus Controller stellt eine Anforderung an den AE
    if BitLow(AECS_E) then

      // ist AECMD_E auf High (inaktiv) so soll ein weiterer
      // Inkrement der Adresse folgen
      if BitHigh(AECMD_E) then

        // Ausgabe der Adresse und INCRQ=aktiv
        PORTC := adr_00_07;
        PORTD := adr_08_15;
        PORTE := adr_16_18;

        NOP;

        // Ausgabe der Adresse ist beendet
        // der BC kann benachrichtigt werden
        ClearBit(AERQ_A); 
        SetBit(LED_GELB_A);

        ASM;

	; Inkrementieren der internen Adresse
        LDS        R16, adr_00_07  ; 1 Cycle
        LDS        R17, adr_08_15  ; 1 Cycle
        LDS        R18, adr_16_18  ; 1 Cycle
        CLR        R19             ; 1 Cycle
        SEC                        ; 1 Cycle
        ADC        R16, R19        ; 1 Cycle
        ADC        R17, R19        ; 1 Cycle
        ADC        R18, R19        ; 1 Cycle
        STS        adr_00_07, R16  ; 1 Cycle
        STS        adr_08_15, R17  ; 1 Cycle
        STS        adr_16_18, R18  ; 1 Cycle
        ; Summe 11 Takte

        ENDASM; 

        // warten bis der BC die Uebernahme quittiert
        while BitLow(AECS_E) do
        endwhile;

        // Abschliessen des Einlesens
        SetBit(AERQ_A);
        ClearBit(LED_GELB_A);


      else

        // PortA => (Eingang)
        DDRA:= %00000000;

        // ist AECMD_E auf Low (aktiv) so soll 
        // ein neuer Befehl uebertragen werden  
        // Einlesen des BefehlsCodes 
        BefehlsCode:=PINA;

        if Befehlscode=INC_ADRESS then

          // der BC kann benachrichtigt werden
          SetBit(LED_GELB_A);
          ClearBit(AERQ_A); 

          // warten bis der BC die Uebernahme quittiert
          while BitLow(AECS_E) do
          endwhile;

          // Abschliessen des Einlesens
          SetBit(AERQ_A);
          ClearBit(LED_GELB_A);

        endif; { if BefehlsCode=ADRESS_TRISTATE then } 
      
      endif; { if BitHigh(AECMD_E) then }

    endif; { if BitLow(AECS_E) then }

  endwhile; { while Befehlscode=INC_ADRESS do }

end PRO_INC_ADRESS;






procedure PRO_ADRESS_TRISTATE;
begin

  // die 3 Adressports werden auf Eingang gesetzt

  // PortC => (Eingang) 
  DDRC:= %00000000;

  // PortD => (Eingang) 
  DDRD:= %00000000;

  // PortE => PE0-PE2 = (3 Adr 16-18) (Ausgang)
  DDRE:= %11111111;
  PortE:=%00000000;


  // der BC kann benachrichtigt werden
  SetBit(LED_GELB_A);
  ClearBit(AERQ_A); 

  // warten bis der BC die Uebernahme quittiert
  while BitLow(AECS_E) do
  endwhile;

  // Abschliessen des Einlesens
  SetBit(AERQ_A);
  ClearBit(LED_GELB_A);


  // die Schleife laeuft so lange der Befehlscode=ADRESS_TRISTATE ist
  while Befehlscode=ADRESS_TRISTATE do

    // der Bus Controller stellt eine Anforderung an den AE
    if BitLow(AECS_E) then

      // ist AECMD_E auf High (inaktiv) so soll ein weiterer
      // Inkrement der Adresse folgen
      if BitLow(AECMD_E) then

        // PortA => (Eingang)
        DDRA:= %00000000;

        // ist AECMD_E auf Low (aktiv) so soll 
        // ein neuer Befehl uebertragen werden  
        // Einlesen des BefehlsCodes 
        BefehlsCode:=PINA;

        if BefehlsCode=ADRESS_TRISTATE then

          // der BC kann benachrichtigt werden
          SetBit(LED_GELB_A);
          ClearBit(AERQ_A); 

          // warten bis der BC die Uebernahme quittiert
          while BitLow(AECS_E) do
          endwhile;

          // Abschliessen des Einlesens
          SetBit(AERQ_A);
          ClearBit(LED_GELB_A);

        endif; { if BefehlsCode=ADRESS_TRISTATE then }  

      else

        SetBit(LED_GELB_A);

      endif; { if BitLow(AECMD_E) then }

    endif; { if BitLow(AECS_E) then }

  endwhile; { while Befehlscode=ADRESS_TRISTATE do }

end PRO_ADRESS_TRISTATE;






procedure PRO_WAIT_Z80ToBuf;
var
  ByteWert : Byte;
begin
  // die 3 Adressports werden auf Eingang gesetzt

  // PortC => (Eingang)
  DDRC:= %00000000;

  // PortD => (Eingang)
  DDRD:= %00000000;

  // PortE => PE0-PE2 = (3 Adr 16-18) (Ausgang)
  DDRE:= %11111111;
  PortE:=%00000000;


  // der BC kann benachrichtigt werden
  SetBit(LED_GELB_A);
  ClearBit(AERQ_A);

  // warten bis der BC die Uebernahme quittiert
  while BitLow(AECS_E) do
  endwhile;

  // Abschliessen des Einlesens
  SetBit(AERQ_A);
  ClearBit(LED_GELB_A);

  ByteBufferCounter:=1;

  // die Schleife laeuft so lange der Befehlscode=ADRESS_TRISTATE ist
  while Befehlscode=WAIT_Z80ToBuf do

    // der Bus Controller stellt eine Anforderung an den AE
    if BitLow(AECS_E) then

      // ist AECMD_E auf High (inaktiv) so soll ein weiterer
      // Inkrement der Adresse folgen
      if BitHigh(AECMD_E) then

       // PortA => (Eingang)
        DDRA:= %00000000;

       // Einlesen Adresse 8..15
        ByteWert:=PinD;
        ByteBuffer[ByteBufferCounter]:=ByteWert;
        Inc(ByteBufferCounter);

       // Einlesen Adresse 0..7
        ByteWert:=PinC;
        ByteBuffer[ByteBufferCounter]:=ByteWert;
        Inc(ByteBufferCounter);

        // Einlesen der Daten
        ByteWert:=PinA;
        ByteBuffer[ByteBufferCounter]:=ByteWert;
        Inc(ByteBufferCounter);

        // der BC kann benachrichtigt werden
        SetBit(LED_GELB_A);
        ClearBit(AERQ_A);

        // warten bis der BC die Uebernahme quittiert
        while BitLow(AECS_E) do
        endwhile;

        // Abschliessen des Einlesens
        SetBit(AERQ_A);
        ClearBit(LED_GELB_A);

      else

        // PortA => (Eingang)
        DDRA:= %00000000;

        // ist AECMD_E auf Low (aktiv) so soll
        // ein neuer Befehl uebertragen werden
        // Einlesen des BefehlsCodes
        BefehlsCode:=PINA;

        if BefehlsCode=WAIT_Z80ToBuf then

          // der BC kann benachrichtigt werden
          SetBit(LED_GELB_A);
          ClearBit(AERQ_A);

          // warten bis der BC die Uebernahme quittiert
          while BitLow(AECS_E) do
          endwhile;

          // Abschliessen des Einlesens
          SetBit(AERQ_A);
          ClearBit(LED_GELB_A);

        endif; { if BefehlsCode=WAIT_Z80ToBuf then }

      endif; { if BitLow(AECMD_E) then }

    endif; { if BitHigh(AECMD_E) then }

  endwhile; { while Befehlscode=WAIT_Z80ToBuf do }

end PRO_WAIT_Z80ToBuf;






procedure PRO_WAIT_BufToBC;
var
  ByteWert : Byte;
begin
  // die 3 Adressports werden auf Eingang gesetzt

  // PortC => (Eingang)
  DDRC:= %00000000;

  // PortD => (Eingang)
  DDRD:= %00000000;

  // PortE => PE0-PE2 = (3 Adr 16-18) (Ausgang)
  DDRE:= %11111111;
  PortE:=%00000000;


  // der BC kann benachrichtigt werden
  SetBit(LED_GELB_A);
  ClearBit(AERQ_A);

  // warten bis der BC die Uebernahme quittiert
  while BitLow(AECS_E) do
  endwhile;

  // Abschliessen des Einlesens
  SetBit(AERQ_A);
  ClearBit(LED_GELB_A);

  ByteBufferCounter:=1;

  // die Schleife laeuft so lange der Befehlscode=ADRESS_TRISTATE ist
  while Befehlscode=WAIT_BufToBC do

    // der Bus Controller stellt eine Anforderung an den AE
    if BitLow(AECS_E) then

      // ist AECMD_E auf High (inaktiv) so soll ein weiterer
      // Inkrement der Adresse folgen
      if BitHigh(AECMD_E) then

        // PortA => (Ausgang)
        DDRA:= %11111111;

        // Einlesen der Daten
        ByteWert:=ByteBuffer[ByteBufferCounter];
        PortA:=ByteWert;
        Inc(ByteBufferCounter);

        // PortA => (Eingang)
        DDRA:= %00000000;

        // der BC kann benachrichtigt werden
        SetBit(LED_GELB_A);
        ClearBit(AERQ_A);

        // warten bis der BC die Uebernahme quittiert
        while BitLow(AECS_E) do
        endwhile;

        // Abschliessen des Einlesens
        SetBit(AERQ_A);
        ClearBit(LED_GELB_A);

      else

        // ist AECMD_E auf Low (aktiv) so soll
        // ein neuer Befehl uebertragen werden
        // Einlesen des BefehlsCodes
        BefehlsCode:=PINA;

        if BefehlsCode=WAIT_BufToBC then

          // der BC kann benachrichtigt werden
          SetBit(LED_GELB_A);
          ClearBit(AERQ_A);

          // warten bis der BC die Uebernahme quittiert
          while BitLow(AECS_E) do
          endwhile;

          // Abschliessen des Einlesens
          SetBit(AERQ_A);
          ClearBit(LED_GELB_A);

        endif; { if BefehlsCode=WAIT_BufToBC then }

      endif; { if BitLow(AECMD_E) then }

    endif; { if BitHigh(AECMD_E) then }

  endwhile; { while Befehlscode=WAIT_BufToBC do }

end PRO_WAIT_BufToBC;



{--------------------------------------------------------------}
{ Hauptprogramm }


begin
  InitPorts;

  adr_00_07 := 0;
  PORTC := adr_00_07;
  
  adr_08_15 := 0;
  PORTD := adr_08_15;
  
  adr_16_18 := 0;
  PORTE := adr_16_18;
  
  Befehlscode := kein_Code;
  
  SetBit(AERQ_A);
  ClearBit(LED_GELB_A);

  loop

    case Befehlscode of

      kein_Code       : PRO_Start;
      |

      LOAD_ADRESS     : PRO_LOAD_ADRESS;
      |

      INC_ADRESS      : PRO_INC_ADRESS;
      |

      ADRESS_TRISTATE : PRO_ADRESS_TRISTATE;
      |

      WAIT_Z80ToBuf   : PRO_WAIT_Z80ToBuf;
      |

      WAIT_BufToBC    : PRO_WAIT_BufToBC;
      |

    else

      SetBit(LED_GELB_A);

    endcase; { case Befehlscode of }
    
  endloop;
  
end AE.

