
// ******************************
//      *** TWI / I2C  ***
// ******************************


// *** Master ***


CONST
{
ProcClock = 22.118.400
100.000 * (16 + 2 * x ) = 3.686.411
16 + (2 * x ) = 36,86411
2 * x = 20,86411
x = 10,432055
x = 11
}

  TWI_BIT_RATE  : Byte = $67;   // fuer 100 KHz
  TWI_PRESCALER : Byte = 0;    // Vorteiler: 0=1 1=4 2=16 3=64

  ADR_DS1307_WR : Byte = %11010000;   // 68H -> D0H
  ADR_DS1307_RD : Byte = %11010001;   // 68H -> D1H


VAR

 TWINT_E[TWCR, 7] : bit; // E


procedure TWI_INIT;
begin
  // TWI - TWBR -> Bit RATE Register
  TWBR := TWI_BIT_RATE;

  // TWI - TWSR -> STATUS-Register
  TWSR := TWI_PRESCALER;

  // TWI Control Register – TWCR
  // Bit 7 – TWINT: TWI Interrupt Flag -> 0
  // Bit 6 – TWEA: TWI Enable Acknowledge Bit -> 1
  // Bit 2 – TWEN: TWI Enable Bit -> 1
//  TWCR := %01000100;

  Error_Code:=0;

end TWI_INIT;




function TWI_Read_Status : Byte;
var
  b1 : Byte;
begin

  b1 := TWSR and %11111000;

  RESULT:=b1;

end TWI_Read_Status;



procedure TWI_Start;
var
  b1 : Byte;
begin

  // Startbit senden

  // TWI Control Register – TWCR
  // Bit 7 – TWINT: TWI Interrupt Flag -> 1
  // Bit 5 – TWSTA: TWI START Condition Bit -> 1
  // Bit 2 – TWEN: TWI Enable Bit -> 1
  TWCR := %10100100;

  // warten bis TWINT 1 wird
  while BitLow(TWINT_E) do
  endwhile;

  b1 := TWSR and %11111000;

  if b1<>$08 then
    Error_Code:=01;
  endif;

end TWI_Start;



procedure TWI_ReStart;
var
  b1 : Byte;
begin

  // Startbit senden

  // TWI Control Register – TWCR
  // Bit 7 – TWINT: TWI Interrupt Flag -> 1
  // Bit 5 – TWSTA: TWI START Condition Bit -> 1
  // Bit 2 – TWEN: TWI Enable Bit -> 1
  TWCR := %10100100;

  // warten bis TWINT 1 wird
  while BitLow(TWINT_E) do
  endwhile;

  b1 := TWSR and %11111000;

  if b1<>$10 then
    Error_Code:=02;
  endif;

end TWI_ReStart;






procedure TWI_Stop;
begin

  // TWI Control Register – TWCR
  // Bit 7 – TWINT: TWI Interrupt Flag -> 1
  // Bit 4 – TWSTO: TWI STOP Condition Bit -> 1
  // Bit 2 – TWEN: TWI Enable Bit -> 1
  TWCR := %10010100;


end TWI_Stop;







procedure TWI_Write(Daten : Byte);
var
  b1 : Byte;
begin

  TWDR:=Daten;

  // Senden der Daten

  // TWI Control Register – TWCR
  // Bit 7 – TWINT: TWI Interrupt Flag -> 1
  // Bit 2 – TWEN: TWI Enable Bit -> 1
  TWCR := %10000100;

  // warten bis TWINT 1 wird
  while BitLow(TWINT_E) do
  endwhile;

end TWI_Write;





function TWI_ReadACK : Byte;
var
  b1 : Byte;
begin

  // Lesen eines Bytes

  // TWI Control Register – TWCR
  // Bit 7 – TWINT: TWI Interrupt Flag -> 1
  // Bit 6 – TWEA: TWI Enable Acknowledge Bit -> 1
  // Bit 2 – TWEN: TWI Enable Bit -> 1
  TWCR := %11000100;

  // warten bis TWINT 1 wird
  while BitLow(TWINT_E) do
  endwhile;

  b1:=TWDR;

  RESULT:=b1;

end TWI_ReadACK;






function TWI_ReadNACK : Byte;
var
  b1 : Byte;
begin

  // Senden der Daten

  // TWI Control Register – TWCR
  // Bit 7 – TWINT: TWI Interrupt Flag -> 1
  // Bit 2 – TWEN: TWI Enable Bit -> 1
  TWCR := %10000100;

  // warten bis TWINT 1 wird
  while BitLow(TWINT_E) do
  endwhile;

  b1:=TWDR;

  RESULT:=b1;

end TWI_ReadNACK;






procedure TWI_1307a;
var
  b1 : Byte;
begin


  TWI_Start;

  // Uebertragung der DS1307 Geraeteadresse und 0 fuer Schreiboperation
  TWI_Write(ADR_DS1307_WR);
  b1:=TWI_Read_Status;
  // Master generiert Start (MT) -> 18H
  if b1<>$18 then
    Error_Code:=03;
  endif;

  // Das erste Register ist das Register0 -> Sekunden
  TWI_Write($00);
  b1:=TWI_Read_Status;
  // Master sendet Daten-Byte, Slave antwortet mit ACK -> 28H
  if b1<>$28 then
    Error_Code:=04;
  endif;

  // Einschreiben von 05 in das Sekunden-Register , CH = 0 = ENABLE
  TWI_Write($05);
  b1:=TWI_Read_Status;
  // Master sendet Daten-Byte, Slave antwortet mit ACK -> 28H
  if b1<>$28 then
    Error_Code:=05;
  endif;

  // Einschreiben von 20 in das Minuten-Register
  TWI_Write($14);
  b1:=TWI_Read_Status;
  // Master sendet Daten-Byte, Slave antwortet mit ACK -> 28H
  if b1<>$28 then
    Error_Code:=06;
  endif;

  // Einschreiben von 18 in das Stunden-Register, Bit7=0 -> 24-Stunden-Anzeige
  TWI_Write($18);
  b1:=TWI_Read_Status;
  // Master sendet Daten-Byte, Slave antwortet mit ACK -> 28H
  if b1<>$28 then
    Error_Code:=07;
  endif;

  TWI_Stop;

end TWI_1307a;







procedure TWI_1307b;
var
  b1 : Byte;
  Sekunden : Byte;
  Minuten : Byte;
  Stunden : Byte;
begin


  TWI_Start;

  // Uebertragung der DS1307 Geraeteadresse und 0 fuer Schreiboperation
  TWI_Write(ADR_DS1307_WR);
  b1:=TWI_Read_Status;
  // Master generiert Start (MT) -> 18H
  if b1<>$18 then
    Error_Code:=08;
  endif;

  // Das erste Register ist das Register0 -> Sekunden
  TWI_Write($00);
  b1:=TWI_Read_Status;
  // Master sendet Daten-Byte, Slave antwortet mit ACK -> 28H
  if b1<>$28 then
    Error_Code:=09;
  endif;

  // Neustart
  TWI_ReStart;

  // Uebertragung der DS1307 Geraeteadresse und 1 fuer Lese-Operation
  TWI_Write(ADR_DS1307_RD);
  b1:=TWI_Read_Status;
  // Master sends slave address (MR), Slave antwortet mit ACK -> 40H
  if b1<>$40 then
    Error_Code:=10;
  endif;

  // Lesen aus dem aktuellen Register (und Inkrement)
  // Lesen der Sekunden
  Sekunden:=TWI_ReadACK;
  b1:=TWI_Read_Status;
  // Master reads data (MR) from Slave, and returns ACK, status code 0x50
  if b1<>$50 then
    Error_Code:=11;
  endif;

  // seconds  =  ((seconds and 0xF0) >> 4)*10 + (seconds and 0x0F)  ' Transform seconds
  {
  b1 := b2 and %01110000;
  b1 := SHR(b1);
  b1 := SHR(b1);
  b1 := SHR(b1);
  b1 := SHR(b1);
  b1 := b1 * 10;
  b2 := b2 and %00001111;
  b2 := b2 + b1;
   }

  // Anzeige der Sekunnden
  Sekunden := Sekunden and %01111111;


  // Lesen aus dem aktuellen Register (und Inkrement)
  // Lesen der Minuten
  Minuten:=TWI_ReadACK;
  b1:=TWI_Read_Status;
  // Master reads data (MR) from Slave, and returns ACK, status code 0x50
  if b1<>$50 then
    Error_Code:=12;
  endif;


  Stunden:=TWI_ReadNACK;
  b1:=TWI_Read_Status;
  // Master reads data (MR) from Slave, and returns NACK, status code 0x58
  if b1<>$58 then
    Error_Code:=13;
  endif;

  TWI_Stop;


  // Anzeige der Uhrzeit
  {
  sende_HEX_Zeichen(Stunden);
  Write_String:=':';
  Write;
  sende_HEX_Zeichen(Minuten);
  Write_String:=':';
  Write;
  sende_HEX_Zeichen_Ln(Sekunden);
                 }

end TWI_1307b;



