

ASM;


; HORIZONTAL SYNC auf LOW
	cbi	  port_control, HSYNC


; Sichern weiterer Register

	push	  zl
	push	  zh
	push	  yl
	push	  yh


; Laden des letzten Standes wichtiger Register
	lds	  yl, Wert_yl
	lds	  yh, Wert_yh


; setzen des Zeigers auf die VLINE-Tabelle
	ldi	  zl,low(2*VLINE)
	ldi	  zh,high(2*VLINE)


; berechnen der aktuellen Tabellen-Position
	add	  zl,yl
	adc	  zh,yh


; lesen des Bytes aus der VLINE-Tabelle in R16
	lpm	  R16,z


; inkrementieren des Adress-Offsets der VLINE-Tabelle
; fuer den naechsten Durchlauf
	adiw	  yh:yl,1


; ist VL_RESET gesetzt, so wird auch der Adress-Offset der
; VLINE-Tabelle auf 0 gesetzt
	sbrc	  R16,VL_RESET
	clr	  yl
	sbrc	  R16,VL_RESET
	clr	  yh


; Abspeichern des Wertes fuer den Adress-Offset
; fuer den naechsten Durchlauf
      	sts	  Wert_yl, yl
	sts	  Wert_yh, yh


; ist VL_VER_IMP gesetzt, so wird VSYNC auf 0 gesetzt
; ansonsten wird es auf 1 gesetzt
	sbrc	  R16, VL_VER_IMP
	cbi	  port_control, VSYNC
	sbrs	  R16, VL_VER_IMP
	sbi	  port_control, VSYNC


; HORIZONTAL SYNC wird auf HIGH gesetzt
	sbi	  port_control, HSYNC


; ist VL_OUT_VA gesetzt, so befinden wir uns im unsichtbaren Teil
; der Anzeige und es wird auf NOVID gesprungen
; Ansonsten geht es im Progeramm weiter
	sbrc	  R16,VL_OUT_VA
        rjmp      VIDEO_END2


; -----------------------------------------------


; ist VL_UP_OFF gesetzt, so werden alle Register fuer die
; Abarbeitung der Text und Grafik Unterprogramme abgespeichert
; und es wird wieder die kleine Entlosschleife geladen
; Ansonsten geht es zu VIDEO_END2

	sbrs	  R16,VL_UP_OFF
	rjmp      VIDEO_END2


; VL_UP_OFF2
; ---------


; ist UP_SCHLEIFE_AKTIV deaktiviert,
; dann wird T3 beendet und T1 aktiviert
;
        lds	  R16, UP_SCHLEIFE_AKTIV
	sbrc	  R16, 0                  ; wenn false (R16=0) dann ...
	rjmp      VIDEO_END2


        ; ... Umschalten T3 und T1


; Interrupt Timer 1 ein
        LDI        R16, 0x02
        STS        TIMSK1, R16

; Interrupt Timer 3 aus
        LDI        R16, 0x00
        STS        TIMSK3, R16



; 1. Aufraumen des Stacks - Sichern aller Register,
;                           die auf den Stack geladen wurden

; 1.1. die auf dem Stack gesicherten Register - aus dem Programm


	pop       yh
	sts	  Save_R29, R29

	pop       yl
	sts	  Save_R28, R28

	pop       zh
	sts	  Save_R31, R31

	pop       zl
	sts	  Save_R30, R30



; 1.2. die auf dem Stack gesicherten Register - aus der ISR

	pop       r16
	sts	  Save_SREG, R16

	pop       r16
	sts	  Save_R16, R16


; 1.3. der auf dem Stack gesicherte PC (2 Byte)

        pop       r16
        sts       Save_SP02, R16

        pop       r16
        sts       Save_SP01, R16


; 2. Sichern der IO-Register und ihrer Werte vom Programm

        in        R16, port_adr_low
        sts       Save_port_adr_low, R16

        in        R16, port_adr_hig
        sts       Save_port_adr_hig, R16

        in        R16, port_dat
        sts       Save_port_dat, R16

        in        R16, ddr_dat
        sts       Save_ddr_dat, R16

        in        R16, port_control
        sts       Save_port_control, R16



; 3. Laden der IO-Register und ihrer Werte fuer Endlosschleife

	lds	  R16, Endlos_port_control
	out       port_control, R16



; Daten-Bus auf Ausgang stellen
	clr       R16
	out       ddr_dat, R16


; 4. Laden des Stacks

        lds	  R16, Endlos_SP01
        push      r16

        lds	  R16, Endlos_SP02
        push      r16

        lds	  R16, Endlos_R16
        push      r16

        lds	  R16, Endlos_SREG
        push      r16


        rjmp      ISR_END2



VIDEO_END2:

	pop       yh
	pop       yl
	pop       zh
	pop       zl


ISR_END2:



ENDASM;



