PROGRAM RSYNTH;

DEVICE = MEGA16;

{$I MEGA16_Register.INC}

{$I mega16_TC.INC}




CONST

  // TaktFrequenz in KHz
  TaktFrequenz : WORD = 16000;

  // bei 16,000 MHz -> 19.200
  UBRR_VAL : Byte = 51;

  // Zuordnung der Kanaele zu den Potis
  POTI1 : Byte = 7;
  POTI2 : Byte = 3;
  POTI3 : Byte = 6;
  POTI4 : Byte = 2;
  POTI5 : Byte = 5;
  POTI6 : Byte = 1;
  POTI7 : Byte = 4;
  POTI8 : Byte = 0;

  Kanal_Counter_Max : Word = 2000;




VAR

  LED_WS[PortD, 6] : bit;  // -> A (High-Aktiv)
  LED_BL[PortD, 7] : bit;  // -> A (High-Aktiv)

  ADC_ADSC[ADCSRA, 6] : bit;  // -> A (High-Aktiv)

  Zaehler : Word;
  SIN_wert : Byte;

  Steuer_String : String[40];
  Sende_String : String[40];

  Kanal_Nr : Byte;

  Wert_POTI1, Wert_POTI2, Wert_POTI3,
  Wert_POTI4, Wert_POTI5, Wert_POTI6,
  Wert_POTI7, Wert_POTI8 : Word;


CONST
{$I SinTab.INC}

{$I SYSTEM.INC}

{$I mega16_ADC.INC}

{$I mega16_UART_oI.INC}



procedure InitPorts;
begin

  {
  A0..A7 - ADC 8 Potentiometer (Eingang)
  }
  DDRA:=  %00000000;
  PortA:= %00000000;

  {
  B0 = SERIAL_DATA (SER) -> 74HC595 (Ausgang)
  B1 = SHIFT_CLOCK (SRCLK) -> 74HC595 (Ausgang)
  B2 = LATCH_CLOCK (RCLK) -> 74HC595 (Ausgang)
  B3 = Tastenfeld 1 -> (Eingang)
  B4 = Tastenfeld 1 -> (Eingang)
  B5 = MOSI
  B6 = MISO
  B7 = SCK
  }
  DDRB:=  %00000111;
  PortB:= %00000000;

  {
  C0..C7 - DAC (Ausgang)
  }
  DDRC:=  %11111111;
  PortC:= %00000000;

  {
  D0 = RXD - UART (Eingang)
  D1 = TXD - UART (Ausgang)
  D2 = frei
  D3 = frei
  D4 = frei
  D5 = frei
  D6 = LED_BL (Ausgang)
  D7 = LED_WS (Ausgang)
  }
  //       LLXXXXTR
  DDRD:=  %11000010;
  PortD:= %00000000;

end InitPorts;





isr TIMER1_COMPA;
begin

ASM;

; Retten der benutzten Register
        PUSH       R17
        PUSH       R18
        PUSH       R19
        PUSH       R24
        PUSH       R25
        PUSH       ZL
        PUSH       ZH

ENDASM;

  SIN_wert:=SINTAB[Zaehler];
  PORTC:=SIN_wert;
  INC(Zaehler);
  if Zaehler=100 then
    Zaehler:=0;
  endif;


ASM;

; Rueckholen der benutzten Register
        POP        ZH
        POP        ZL
        POP        R25
        POP        R24
        POP        R19
        POP        R18
        POP        R17

ENDASM;


end TIMER1_COMPA;






procedure ADC_senden;
begin

  WordToString(Wert_POTI1);
  Sende_String:='d'+Write_String_Word;

  WordToString(Wert_POTI2);
  Sende_String:=Sende_String+Write_String_Word;

  WordToString(Wert_POTI3);
  Sende_String:=Sende_String+Write_String_Word;

  WordToString(Wert_POTI4);
  Sende_String:=Sende_String+Write_String_Word;

  WordToString(Wert_POTI5);
  Sende_String:=Sende_String+Write_String_Word;

  WordToString(Wert_POTI6);
  Sende_String:=Sende_String+Write_String_Word;

  WordToString(Wert_POTI7);
  Sende_String:=Sende_String+Write_String_Word;

  WordToString(Wert_POTI8);
  Sende_String:=Sende_String+Write_String_Word;

  Write_String:=Sende_String;

  WriteLn;

end ADC_senden;






procedure Steuer_String_auswerten;
var
  Buchstabe : Char;
  Zeichen_1 : Char;
  Zeichen_2 : Char;
  Laenge : Byte;
begin
  Laenge:=Length(Steuer_String);

  if Laenge>0 then
    Buchstabe:=Steuer_String[1];
    case Buchstabe of

            // a
      'a' : if Laenge=1 then
              SetBit(LED_BL);
            else
              Write_String:='x01';
              WriteLn;
            endif;

      |


           // b
      'b' : if Laenge=1 then
              ClearBit(LED_BL);
            else
              Write_String:='x01';
              WriteLn;
            endif;

      |


           // c
      'c' : if Laenge=1 then
              ADC_senden;
            else
              Write_String:='x01';
              WriteLn;
            endif;

      |


    else
      Write_String:='x02';
      WriteLn;
    endcase;

  endif; { if Laenge>0 then }

end Steuer_String_auswerten;






procedure Steuer_String_lesen;
var
  akt_Zeichen : Char;
begin
  if Zeichen<>$0D then // 0D wird ignoriert

    if Zeichen=$0A then  // bei 0A wird ausgerwertet
      Steuer_String_auswerten;
      Steuer_String:='';
    else
      akt_Zeichen:=ByteToChar(Zeichen);
      Steuer_String:= Steuer_String + akt_Zeichen; // sonst nur abgespeichert
    endif; { if akt_Zeichen<>char($0A) }

  endif; { if akt_Zeichen<>char($0D) then }

end Steuer_String_lesen;








BEGIN

  // Initialisierung der Ports
  InitPorts;

  // Initialisierung der seriellen Schnittstelle
  INIT_SER_PORT;

  // Initialisierung des AD-Wandlers
  Init_ADC;


  ClearBit(LED_BL);
  ClearBit(LED_WS);



  Zaehler:=0;

  // Frequenz = 16.000.000 Hz;
  // Frequenz / 100 = 160.000 Hz bei 100 Samles fuer eine Periode
  // 182 = 880 Hz (440 Ton a)
  // 160 = 1 KHz
  //  80 = 2 KHz
//  Init_TIMER1_COMPA(80);


//  EnableInts;


  Kanal_Nr:=POTI1;
  ADC_starten(POTI1);


  loop


    RX_Puffer_abfragen;

    if Zeichen>0 then
      Steuer_String_lesen;
    endif;


    case Kanal_Nr of

      POTI1 : if ADC_bereit then
                Wert_POTI1:=ADC_lesen;
                Kanal_Nr:=POTI2;
                ADC_starten(POTI2);
              endif;
      |

      POTI2 : if ADC_bereit then
                Wert_POTI2:=ADC_lesen;
                Kanal_Nr:=POTI3;
                ADC_starten(POTI3);
              endif;
      |

      POTI3 : if ADC_bereit then
                Wert_POTI3:=ADC_lesen;
                Kanal_Nr:=POTI4;
                ADC_starten(POTI4);
              endif;
      |

      POTI4 : if ADC_bereit then
                Wert_POTI4:=ADC_lesen;
                Kanal_Nr:=POTI5;
                ADC_starten(POTI5);
              endif;
      |

      POTI5 : if ADC_bereit then
                Wert_POTI5:=ADC_lesen;
                Kanal_Nr:=POTI6;
                ADC_starten(POTI6);
              endif;
      |

      POTI6 : if ADC_bereit then
                Wert_POTI6:=ADC_lesen;
                Kanal_Nr:=POTI7;
                ADC_starten(POTI7);
              endif;
      |

      POTI7 : if ADC_bereit then
                Wert_POTI7:=ADC_lesen;
                Kanal_Nr:=POTI8;
                ADC_starten(POTI8);
              endif;
      |

      POTI8 : if ADC_bereit then
                Wert_POTI8:=ADC_lesen;
                Kanal_Nr:=POTI1;
                ADC_starten(POTI1);
              endif;
      |

    endcase;

  endloop;



END RSYNTH.
